#! /usr/bin/env python
#
"""
Utility classes and methods for dealing with URLs.
"""
# Copyright (C) 2011, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'


import os

import gc3libs.compat.urlparse as urlparse


# XXX: rewrite using `collections.namedtuple` when we no longer
# support 2.4 and 2.5
class Url(tuple):
    """
    Represent a URL as a named-tuple object.  This is an immutable
    object that cannot be changed after creation.  

    The following read-only attributes are defined on objects of class `Url`.
    
    =========   =====   ===================================  ====================
    Attribute   Index   Value                                Value if not present
    =========   =====   ===================================  ====================
    scheme      0       URL scheme specifier                 empty string
    netloc      1       Network location part                empty string
    path        2       Hierarchical path                    empty string
    query       3       Query component                      empty string
    hostname    4       Host name (lower case)               None
    port        5       Port number as integer (if present)  None
    username    6       User name                            None
    password    7       Password                             None
    =========   =====   ===================================  ====================

    There are two ways of constructing `Url` objects:

    * By passing a string `urlstring`::

        >>> u = Url('http://www.example.org/data')

        >>> u.scheme
        'http'
        >>> u.netloc
        'www.example.org'
        >>> u.path
        '/data'

      The default URL scheme is ``file``::

        >>> u = Url('/tmp/foo')
        >>> u.scheme
        'file'
        >>> u.path
        '/tmp/foo'

      Please note that extra leading slashes '/' are interpreted as
      the begining of a network location:

        >>> u = Url('//foo/bar')
        >>> u.path
        '/bar'
        >>> u.netloc
        'foo'
        >>> Url('///foo/bar').path
        '/foo/bar'

      Check RFC 3986 http://tools.ietf.org/html/rfc3986

      If `force_abs` is `True` (default), then the `path`
      attribute is made absolute, by calling `os.path.abspath` if
      necessary::

        >>> u = Url('foo/bar', force_abs=True)
        >>> os.path.isabs(u.path)
        True

      Otherwise, if `force_abs` is `False`, then the `path`
      attribute stores the passed string unchanged::

        >>> u = Url('foo', force_abs=False)
        >>> os.path.isabs(u.path)
        False
        >>> u.path
        'foo'

      Other keyword arguments can specify defaults for missing parts
      of the URL::

        >>> u = Url('/tmp/foo', scheme='file', netloc='localhost')
        >>> u.scheme
        'file'
        >>> u.netloc
        'localhost'
        >>> u.path
        '/tmp/foo'

    * By passing keyword arguments only, to construct an `Url` object
      with exactly those values for the named fields::

        >>> u = Url(scheme='http', netloc='www.example.org', path='/data')

      In this form, the `force_abs` parameter is ignored.

    See also: http://docs.python.org/library/urlparse.html#urlparse-result-object
    """
    __slots__ = ( )

    _fields = ['scheme', 'netloc', 'path', 'hostname', 'port', 'username', 'password']

    def __new__(cls, urlstring=None, force_abs=True,
                scheme='file', netloc='', path='',
                hostname=None, port=None, username=None, password=None):
        """
        Create a new `Url` object.  See the `Url`:class: documentation
        for invocation syntax.
        """
        if urlstring is not None:
            if isinstance(urlstring, Url):
                # copy constructor
                return tuple.__new__(cls, (
                    urlstring.scheme, urlstring.netloc, urlstring.path,
                    urlstring.hostname, urlstring.port,
                    urlstring.username, urlstring.password,
                    ))
            else:
                # parse `urlstring` and use kwd arguments as default values
                try:
                    urldata = urlparse.urlsplit(urlstring, scheme=scheme, allow_fragments=False)
                    if urldata.scheme == 'file' and not os.path.isabs(urldata.path) and force_abs:
                        urldata = urlparse.urlsplit('file://' + os.path.abspath(urldata.path))
                    return tuple.__new__(cls, (
                        urldata.scheme or scheme,
                        urldata.netloc or netloc,
                        urldata.path or path,
                        urldata.hostname or hostname,
                        urldata.port or port,
                        urldata.username or username,
                        urldata.password or password,
                        ))
                except (ValueError, TypeError, AttributeError), ex:
                    raise ValueError("Cannot parse string '%s' as a URL: %s: %s"
                                     % (urlstring, ex.__class__.__name__, str(ex)))
        else:
            # no `urlstring`, use kwd arguments
            return tuple.__new__(cls, (
                scheme, netloc, path,
                hostname, port,
                username, password,
                ))


    def __getattr__(self, name):
        try:
            return self[self._fields.index(name)]
        except ValueError:
            raise AttributeError("'%s' object has no attribute '%s'"
                                 % (self.__class__.__name__, name))
    

    def __getnewargs__(self):
        """Support pickling/unpickling `Url` class objects."""
        return (None, False, # urlstring, force_abs
                self.scheme, self.netloc, self.path, self.hostname, self.port, self.username, self.password)


    def __repr__(self):
        """
        Return a printed representation of this object, such that `eval(repr(x)) == x`.
        """
        return ("Url(scheme=%r, netloc=%r, path=%r, hostname=%r, port=%r, username=%r, password=%r)"
                % (self.scheme, self.netloc, self.path, self.hostname, self.port, self.username, self.password))


    def __str__(self):
        """
        Return a URL string corresponding to the URL object `urldata`.

            >>> u = Url('gsiftp://gridftp.example.org:2811/data')
            >>> str(u)
            'gsiftp://gridftp.example.org:2811/data'

        If the URL was constructed by parsing a URL string, a
        different string can result, although equivalent according to
        the URL definition::

            >>> u1 = Url('/tmp/foo')
            >>> str(u1)
            'file:///tmp/foo'

        """
        url = self.path
        # XXX: assumes all our URLs are of netloc-type!
        if self.netloc or (self.scheme and not url.startswith('//')):
            if url and not url.startswith('/'):
                url = '/' + url
            url = '//' + (self.netloc or '') + url
        if self.scheme:
            url = self.scheme + ':' + url
        return url    


    def __eq__(self, other):
        """
        Return `True` if `self` and `other` are equal `Url` objects,
        or if their string representations match.

        Examples:

          >>> u = Url('/tmp/foo')
          >>> u == Url('/tmp/foo')
          True
          >>> u == Url('file:///tmp/foo')
          True
          >>> u == Url('http://example.org/')
          False
          
          >>> u == str(u)
          True
          >>> u == '/tmp/foo'
          True
          >>> u == 'file:///tmp/foo'
          True
          >>> u == 'http://example.org'
          False

          >>> u == 42
          False
        
        """
        try:
            # The `tuple.__eq__` call can only be used if both `self`
            # and `other` are `tuple` subclasses; we know that `self`
            # is, but we need to check `other`.
            return ((isinstance(other, tuple) and tuple.__eq__(self, other))
                    or str(self) == str(other)
                    or tuple.__eq__(self, Url(other)))
        except ValueError, ex:
            # `other` is not a URL and cannot be made into one
            return False
    
    def __ne__(self, other):
        """
        The opposite of `__eq__`.
        """
        return not self.__eq__(other)
    
    
    def adjoin(self, relpath):
        """
        Return a new `Url`, constructed by appending `relpath` to the
        path section of this URL.

        Example::

            >>> u0 = Url('http://www.example.org')
            >>> u1 = u0.adjoin('data')
            >>> str(u1)
            'http://www.example.org/data'

            >>> u2 = u1.adjoin('moredata')
            >>> str(u2)
            'http://www.example.org/data/moredata'

        Even if `relpath` starts with `/`, it is still appended to the
        path in the base URL::

            >>> u3 = u2.adjoin('/evenmore')
            >>> str(u3)
            'http://www.example.org/data/moredata/evenmore'
        
        """
        if relpath.startswith('/'):
            relpath = relpath[1:]
        return Url(scheme=self.scheme, netloc=self.netloc,
                   path=os.path.join((self.path or '/'), relpath),
                   hostname=self.hostname, port=self.port,
                   username=self.username, password=self.password)
    

class UrlKeyDict(dict):
    """
    A dictionary class enforcing that all keys are URLs.

    Strings and/or objects returned by `urlparse` can
    be used as keys.  Setting a string key automatically
    translates it to a URL:

       >>> d = UrlKeyDict()
       >>> d['/tmp/foo'] = 1
       >>> for k in d.keys(): print (type(k), k.path) # doctest:+ELLIPSIS
       (<class '....Url'>, '/tmp/foo')

    Retrieving the value associated with a key works with both the
    string or the url value of the key:

        >>> d['/tmp/foo']
        1
        >>> d[Url('/tmp/foo')]
        1

    Key lookup can use both the string or the `Url` value as well:

        >>> '/tmp/foo' in d
        True
        >>> Url('/tmp/foo') in d
        True
        >>> 'file:///tmp/foo' in d
        True
        >>> 'http://example.org' in d
        False

    Class `UrlKeyDict` supports initialization by copying items from
    another `dict` instance or from an iterable of (key, value)
    pairs::

        >>> d1 = UrlKeyDict({ '/tmp/foo':'foo', '/tmp/bar':'bar' })
        >>> d2 = UrlKeyDict([ ('/tmp/foo', 'foo'), ('/tmp/bar', 'bar') ])
        >>> d1 == d2
        True

    Differently from `dict`, initialization from keyword arguments
    alone is *not* supported:

        >>> d3 = UrlKeyDict(foo='foo', bar='bar') # doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        TypeError: __init__() got an unexpected keyword argument 'foo'


    An empty `UrlKeyDict` instance is returned by the constructor
    when called with no parameters::

        >>> d0 = UrlKeyDict()
        >>> len(d0)
        0

    If `force_abs` is `True`, then all paths are converted to
    absolute ones in the dictionary keys.

        >>> d = UrlKeyDict(force_abs=True)
        >>> d['foo'] = 1
        >>> for k in d.keys(): print os.path.isabs(k.path)
        True

        >>> d = UrlKeyDict(force_abs=False)
        >>> d['foo'] = 2
        >>> for k in d.keys(): print os.path.isabs(k.path)
        False

    """
    def __init__(self, iter_or_dict=None, force_abs=False):
        self._force_abs = force_abs
        if iter_or_dict is not None:
            try:
                # if `iter_or_dict` is a dict-like object, then it has `iteritems()`
                for k,v in iter_or_dict.iteritems():
                    self[k] = v
            except AttributeError:
                # then assume `iter_or_dict` is an iterator over (key, value) pairs
                for k,v in iter_or_dict:
                    self[k] = v

    def __contains__(self, key):
        # this is necessary to have key-lookup work with strings as well
        return (dict.__contains__(self, key)
                or key in self.keys())

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError, ex:
            # map `key` to a URL and try with that
            try:
                return dict.__getitem__(self, Url(key, self._force_abs))
            except:
                raise ex

    def __setitem__(self, key, value):
        try:
            dict.__setitem__(self, Url(key, self._force_abs), value)
        except:
            dict.__setitem__(self, key, value)


class UrlValueDict(dict):
    """
    A dictionary class enforcing that all values are URLs.

    Strings and/or objects returned by `urlparse` can
    be used as values.  Setting a string value automatically
    translates it to a URL:

       >>> d = UrlValueDict()
       >>> d[1] = '/tmp/foo'
       >>> d[2] = Url('file:///tmp/bar')
       >>> for v in d.values(): print (type(v), v.path) # doctest:+ELLIPSIS
       (<class '....Url'>, '/tmp/foo')
       (<class '....Url'>, '/tmp/bar')

    Retrieving the value associated with a key always returns the
    URL-type value, regardless of how it was set::

        >>> repr(d[1])
        "Url(scheme='file', netloc='', path='/tmp/foo', hostname=None, port=None, username=None, password=None)"

    Class `UrlValueDict` supports initialization by any of the
    methods that work with a plain `dict` instance::

        >>> d1 = UrlValueDict({ 'foo':'/tmp/foo', 'bar':'/tmp/bar' })
        >>> d2 = UrlValueDict([ ('foo', '/tmp/foo'), ('bar', '/tmp/bar') ])
        >>> d3 = UrlValueDict(foo='/tmp/foo', bar='/tmp/bar')

        >>> d1 == d2
        True
        >>> d2 == d3
        True

    In particular, an empty `UrlDict` instance is returned by the
    constructor when called with no parameters::

        >>> d0 = UrlValueDict()
        >>> len(d0)
        0

    If `force_abs` is `True`, then all paths are converted to
    absolute ones in the dictionary values.

        >>> d = UrlValueDict(force_abs=True)
        >>> d[1] = 'foo'
        >>> for v in d.values(): print os.path.isabs(v.path)
        True

        >>> d = UrlValueDict(force_abs=False)
        >>> d[2] = 'foo'
        >>> for v in d.values(): print os.path.isabs(v.path)
        False

    """
    def __init__(self, iter_or_dict=None, force_abs=False, **extra_args):
        self._force_abs = force_abs
        if iter_or_dict is not None:
            try:
                # if `iter_or_dict` is a dict-like object, then it has `iteritems()`
                for k,v in iter_or_dict.iteritems():
                    self[k] = v
            except AttributeError:
                # then assume `iter_or_dict` is an iterator over (key, value) pairs
                for k,v in iter_or_dict:
                    self[k] = v
        for k,v in extra_args.iteritems():
            self[k] = v

    def __setitem__(self, key, value):
        try:
            dict.__setitem__(self, key, Url(value, self._force_abs))
        except:
            dict.__setitem__(self, key, value)



## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="url",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
Collection of tools to supplement optimization algorithm 
:class:`optimizer.EvolutionaryAlgorithm <gc3libs.optimizer.EvolutionaryAlgorithm>`. 

Include a list of desired tools in param `after_update_opt_state` of
:class:`optimizer.EvolutionaryAlgorithm <gc3libs.optimizer.EvolutionaryAlgorithm>`. 
"""
# Copyright (C) 2011, 2012, 2013 University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__version__ = '$Revision$'
__author__ = 'Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
__docformat__ = 'reStructuredText'


import os
import sys
import logging

log = logging.getLogger("gc3.gc3libs")

class plot_population(object):
    '''Plot the 2-dimensional population of an 
       :class:`gc3libs.optimizer.EvolutionaryAlgorithm` instance. 
       If the population is not 2-d an error message appears and no
       plot is created. 
       
       :param str figure_dir: Path to the directory where plots should be stored. 
                              Directory will be created if non-existent. 
    '''
    def __init__(self, figure_dir):
        if not os.path.isdir(figure_dir):
            os.mkdir(figure_dir)        
        self.figure_dir = figure_dir

    def __call__(self, algo):
        '''Plot population for `algo`.  
        
           :param str algo: Instance of :class:`gc3libs.optimizer.EvolutionaryAlgorithm`. 
        '''        
        log.debug('entering plot_population.__call__')
        pop = algo.pop
        if not algo.dim == 2:
            algo.logger.critical('plot_population is implemented only for algo.dim = 2')
            return
        import matplotlib
        matplotlib.use('SVG')
        import matplotlib.pyplot as plt
        x = pop[:, 0]
        y = pop[:, 1]
        # determine bounds
        x0_min = algo.pop[:,0].min()
        x1_min = algo.pop[:,1].min()
        x0_max = algo.pop[:,0].max()
        x1_max = algo.pop[:,1].max()
        
        # Determine plotting area
        x_dif = x0_max - x0_min
        y_dif = x1_max - x1_min
        scaleFac = 0.3
        x0_min_plot = x0_min - scaleFac * x_dif
        x0_max_plot = x0_max + scaleFac * x_dif
        x1_min_plot = x1_min - scaleFac * y_dif
        x1_max_plot = x1_max + scaleFac * y_dif
    
        # make plot
        fig = plt.figure()
        ax = fig.add_subplot(111)
    
        ax.scatter(x, y)
        # x box constraints
        ax.plot([x0_min, x0_min], [x1_min_plot, x1_max_plot])
        ax.plot([x0_max, x0_max], [x1_min_plot, x1_max_plot])
        ax.plot([x0_min_plot, x0_max_plot], [x1_min, x1_min])
        ax.plot([x0_min_plot, x0_max_plot], [x1_max, x1_max])
        
        ## all other linear constraints
        #c_x_min_plot = algo.filter_fn.linearConstr(x_min_plot)
        #c_xmax = algo.filter_fn.linearConstr(xmax)
        #for ixC in range(len(c_x_min_plot)):
            #ax.plot([x_min_plot, xmax], [c_x_min_plot[ixC], c_xmax[ixC]])
        ax.axis(xmin = x0_min_plot, xmax = x0_max_plot,
                ymin = x1_min_plot, ymax = x1_max_plot)
        
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        
        ax.set_title('Best: x %s, f(x) %f' % (algo.best_x, algo.best_y))

        fig.savefig(os.path.join(self.figure_dir, 'pop%d' % (algo.cur_iter)))
    
def print_stats(algo, output=sys.stdout):
    '''Print summary statistics for `algo`.  
    
       :param str algo: Instance of :class:`gc3libs.optimizer.EvolutionaryAlgorithm`. 
       :param output: Output stream. 
    '''
    output.write('Iteration: %d,  x: %s f(x): %f\n' % (algo.cur_iter, algo.best_x, algo.best_y))

def log_stats(algo, logger=logging.getLogger()):
    '''Log summary statistics for `algo`.  
    
       :param str algo: Instance of :class:`gc3libs.optimizer.EvolutionaryAlgorithm`. 
    '''
    logger.info('Iteration: %d,  x: %s f(x): %f',
                algo.cur_iter, algo.best_x, algo.best_y)



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011, 2012, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

import datetime
import os
import tempfile

from nose.tools import assert_equal

import gc3libs
import gc3libs.core
import gc3libs.config
from gc3libs.quantity import Memory, kB, MB, GB, Duration, seconds, minutes, hours

State = gc3libs.Run.State

from faketransport import FakeTransport

files_to_remove = []

def correct_submit(jobid=123):
    out = """Your job %s ("DemoSGEApp") has been submitted
""" % jobid
    return (0, out, "")

def correct_qstat_queued(jobid=123):
    out = """  %s 0.00000 DemoSGEApp antonio      qw    03/15/2012 08:53:34                                    1
""" % jobid
    return (0, out, "")


def correct_qstat_running(jobid=123):
    out = """  %s 0.55500 DemoSGEApp antonio      r     03/15/2012 08:53:42 all.q@compute-0-2.local            1
""" % jobid
    return (0, out, "")

def qacct_notfound(jobid=123):
    err = """error: job id %s not found
""" % jobid
    return (1, "", err)

def qdel_notfound(jobid=123):
    return (1,"",  """denied: job "%s" does not exist""" % jobid)

def qstat_notfound(jobid=123):
    return (1, "", "")

def correct_qacct_done(jobid=123):
    out = """
==============================================================
qname        all.q
hostname     compute-0-2.local
group        antonio
owner        antonio
project      NONE
department   defaultdepartment
jobname      DemoSGEApp
jobnumber    %s
taskid       undefined
account      sge
priority     0
qsub_time    Thu Mar 15 08:42:46 2012
start_time   Thu Mar 15 08:43:00 2012
end_time     Thu Mar 15 08:43:10 2012
granted_pe   NONE
slots        1
failed       0
exit_status  0
ru_wallclock 10
ru_utime     0.154
ru_stime     0.094
ru_maxrss    0
ru_ixrss     0
ru_ismrss    0
ru_idrss     0
ru_isrss     0
ru_minflt    22296
ru_majflt    0
ru_nswap     0
ru_inblock   0
ru_oublock   0
ru_msgsnd    0
ru_msgrcv    0
ru_nsignals  0
ru_nvcsw     306
ru_nivcsw    157
cpu          0.248
mem          0.000
io           0.006
iow          0.000
maxvmem      13.152M
arid         undefined
""" % jobid

    return (0, out, "")


def qsub_failed_jobnamestartswithdigit(jobname='123DemoSGEApp'):
    out = ""
    err = """Unable to run job: denied: "%s" is not a valid object name (cannot start with a digit).
Exiting.
""" % jobname
    return (1, out, err)

# def qsub_failed_acl():
#     out = ""
#     err = """qsub: Unauthorized Request  MSG=user ACL rejected the submitting user: user amessina@argo.ictp.it, queue cm1
# """
#     return (159, out, err)

def qdel_success(jobid=123):
    return (0, "antonio has registered the job %s for deletion" % jobid, "")

def qdel_failed_acl(jobid=123):
    err = """antonio - you do not have the necessary privileges to delete the job "%s"
""" % jobid
    out = ""
    return (1, out, err)


class FakeApp(gc3libs.Application):
    def __init__(self, **extra_args):
        gc3libs.Application.__init__(
            self,
            arguments = ['/bin/hostname'], # mandatory
            inputs = [],                  # mandatory
            outputs = [],                 # mandatory
            output_dir = "./fakedir",    # mandatory
            stdout = "stdout.txt",
            stderr = "stderr.txt",
            requested_cores = 1, **extra_args)


class TestBackendSge(object):

    CONF="""
[resource/example]
type=sge
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64
queue=testing
enabled=True

[auth/ssh]
type=ssh
username=NONEXISTENT
"""
    def setUp(self):
        (fd, self.tmpfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        f.write(TestBackendSge.CONF)
        f.close()

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(self.tmpfile)

        self.core = gc3libs.core.Core(self.cfg)

        self.backend = self.core.get_backend('example')
        self.backend.transport = FakeTransport()
        self.transport = self.backend.transport
        # Basic responses
        self.transport.expected_answer['qstat'] = qstat_notfound()
        self.transport.expected_answer['tracejob'] = qacct_notfound()
        self.transport.expected_answer['qdel'] = qdel_notfound()

    def tearDown(self):
        os.remove(self.tmpfile)

    def test_submission_failed(self):
        app = FakeApp()
        # Submission failed (jobname starting with a digit, cfr issue #250
        # at http://code.google.com/p/gc3pie/issues/detail?id=250) This
        # first test will show the answer you would get if the job name
        # starts with a digit.
        self.transport.expected_answer['qsub'] = qsub_failed_jobnamestartswithdigit()
        try:
            self.core.submit(app)
            assert False
        except Exception, e:
            assert isinstance(e, gc3libs.exceptions.LRMSError)
        assert_equal(app.execution.state, State.NEW)

        # This second example will show how the Application.__init__()
        # method changes the jobname in order not to have a digit at the
        # beginning of it.
        app = FakeApp(jobname='123Demo')
        assert_equal(app.jobname, 'GC3Pie.123Demo')
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        assert_equal(app.execution.state, State.SUBMITTED)


    #     # Submission failed (unauthrozed user):
    #     t.expected_answer['qsub'] = qsub_failed_acl()
    #     try:
    #         self.core.submit(app)
    #     except Exception, e:
    #         assert isinstance(e, gc3libs.exceptions.LRMSError)
    #     assert_equal(app.execution.state, State.NEW)


    def test_sge_basic_workflow(self):
        app = FakeApp()
        # Succesful submission:
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        assert_equal(app.execution.state, State.SUBMITTED)


        # Update state. We would expect the job to be SUBMITTED
        self.transport.expected_answer['qstat'] = correct_qstat_queued()
        self.transport.expected_answer['qacct'] = qacct_notfound()
        self.core.update_job_state(app)
        assert_equal(app.execution.state, State.SUBMITTED)

        # Update state. We would expect the job to be RUNNING
        self.transport.expected_answer['qstat'] = correct_qstat_running()
        self.transport.expected_answer['qacct'] = qacct_notfound()
        self.core.update_job_state(app)
        assert_equal(app.execution.state, State.RUNNING)

        # Job done. qstat doesn't find it, qacct should.
        self.transport.expected_answer['qstat'] = qstat_notfound()
        self.transport.expected_answer['qacct'] = correct_qacct_done()
        self.core.update_job_state(app)
        assert_equal(app.execution.state, State.TERMINATING)


    def test_qacct_parsing(self):
        app = FakeApp()
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        self.transport.expected_answer['qacct'] = correct_qacct_done()
        self.core.update_job_state(app)
        assert_equal(app.execution.state, State.TERMINATING)

        job = app.execution
        # common job reporting values (see Issue 78)
        assert_equal(job.exitcode,        0)
        assert_equal(job.returncode,      0)
        assert_equal(job.duration,        10*seconds)
        assert (job.max_used_memory - 13.152*MB) < 1*MB # floating point approx
        assert_equal(job.used_cpu_time,   0.248*seconds)
        # SGE-specific values
        assert_equal(job.sge_queue,       'all.q')
        assert_equal(job.sge_jobname,     'DemoSGEApp')
        assert_equal(job.sge_submission_time,
                     datetime.datetime(year=2012, month=3, day=15, hour=8, minute=42, second=46))
        assert_equal(job.sge_start_time,
                     datetime.datetime(year=2012, month=3, day=15, hour=8, minute=43, second=0))
        assert_equal(job.sge_completion_time,
                     datetime.datetime(year=2012, month=3, day=15, hour=8, minute=43, second=10))
        assert_equal(job.sge_failed,      0)

    def test_delete_job(self):
        app = FakeApp()
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)

        self.transport.expected_answer['qdel'] = qdel_success()
        self.core.kill(app)
        assert_equal(app.execution.state, State.TERMINATED)

    def test_delete_job(self):
        app = FakeApp()
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        assert_equal(app.execution.state, State.SUBMITTED)

        self.transport.expected_answer['qdel'] = qdel_failed_acl()
        self.core.kill(app)
        assert_equal(app.execution.state, State.TERMINATED)


def tearDownModule():
    for fname in files_to_remove:
        if os.path.isdir(fname):
            shutil.rmtree(fname)
        else:
            os.remove(fname)


def test_get_command():
    (fd, tmpfile) = tempfile.mkstemp()
    files_to_remove.append(tmpfile)
    f = os.fdopen(fd, 'w+')
    f.write("""
[auth/ssh]
type=ssh
username=NONEXISTENT

[resource/example]
# mandatory stuff
type=sge
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64

# alternate command paths
qsub = /usr/local/bin/qsub -q testing
qacct = /usr/local/sbin/qacct
qstat = /usr/local/bin/qstat
qdel = /usr/local/bin/qdel # comments are ignored!
""")
    f.close()

    cfg = gc3libs.config.Configuration()
    cfg.merge_file(tmpfile)
    b = cfg.make_resources()['example']

    assert_equal(b.qsub, ['/usr/local/bin/qsub', '-q', 'testing'])

    assert_equal(b._qacct, '/usr/local/sbin/qacct')
    assert_equal(b._qdel,  '/usr/local/bin/qdel')
    assert_equal(b._qstat, '/usr/local/bin/qstat')


if __name__ == "__main__":
    import nose
    nose.runmodule()



#! /usr/bin/env python
#
"""
Tools for debugging GC3Libs based programs.

Part of the code used in this module originally comes from:
  - http://wordaligned.com/articles/echo

"""
# Copyright (C) 2011 GC3, University of Zurich, except where otherwise noted.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'


import inspect
import itertools
import sys

import gc3libs.compat.functools as functools

import gc3libs


def name(item):
    """Return an item's name."""
    return item.__name__

    
def is_classmethod(instancemethod):
    """Determine if an instancemethod is a classmethod."""
    return instancemethod.im_self is not None

    
def is_class_private_name(name):
    """Determine if a name is a class private name."""
    # Exclude system defined names such as __init__, __add__ etc
    return name.startswith("__") and not name.endswith("__")

    
def method_name(method):
    """
    Return a method's name.
    
    This function returns the name the method is accessed by from
    outside the class (i.e. it prefixes "private" methods appropriately).
    """
    mname = name(method)
    if is_class_private_name(mname):
        mname = "_%s%s" % (name(method.im_class), mname)
    return mname


def format_arg_value(arg_val):
    """
    Return a string representing a (name, value) pair.

    Example::

      >>> format_arg_value(('x', (1, 2, 3)))
      'x=(1, 2, 3)'
    """
    arg, val = arg_val
    return "%s=%r" % (arg, val)


def trace(fn, log=gc3libs.log.debug):
    """
    Logs calls to a function.
    
    Returns a decorated version of the input function which "echoes" calls
    made to it by writing out the function's name and the arguments it was
    called with.
    """
    # Unpack function's arg count, arg names, arg defaults
    code = fn.func_code
    argcount = code.co_argcount
    argnames = code.co_varnames[:argcount]
    fn_defaults = fn.func_defaults or list()
    argdefs = dict(zip(argnames[-len(fn_defaults):], fn_defaults))
    
    @functools.wraps(fn)
    def wrapped(*v, **k):
        # Collect function arguments by chaining together positional,
        # defaulted, extra positional and keyword arguments.
        positional = [ format_arg_value(nm_val)
                       for nm_val in zip(argnames, v) ]
        defaulted = [ format_arg_value((a, argdefs[a]))
                      for a in argnames[len(v):] if a not in k ]
        nameless = [ repr(val) for val in v[argcount:] ]
        keyword = [ format_arg_value((key,val))
                    for key,val in k.items() ]
        args = positional + defaulted + nameless + keyword
        log("%s(%s)" % (name(fn), str.join(", ", args)))
        return fn(*v, **k)
    return wrapped


def trace_instancemethod(cls, method, log=gc3libs.log.debug):
    """
    Change an instancemethod so that calls to it are traced.
    
    Replacing a classmethod is a little more tricky.
    See: http://www.python.org/doc/current/ref/types.html
    """
    mname = method_name(method)
    never_echo = "__str__", "__repr__", # Avoid recursion printing method calls
    if mname in never_echo:
        pass
    elif is_classmethod(method):
        setattr(cls, mname, classmethod(trace(method.im_func, log)))
    else:
        setattr(cls, mname, trace(method, log))


def trace_class(cls, log=gc3libs.log.debug):
    """
    Trace calls to class methods and static functions
    """
    for _, method in inspect.getmembers(cls, inspect.ismethod):
        trace_instancemethod(cls, method, log)
    for _, fn in inspect.getmembers(cls, inspect.isfunction):
        setattr(cls, name(fn), staticmethod(trace(fn, log)))

    
def trace_module(mod, log=gc3libs.log.debug):
    """
    Trace calls to functions and methods in a module.
    """
    for fname, fn in inspect.getmembers(mod, inspect.isfunction):
        setattr(mod, fname, trace(fn, log))
    for _, cls in inspect.getmembers(mod, inspect.isclass):
        trace_class(cls, log)



## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="debug",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
"""
Interface to different resource management systems for the GC3Libs.
"""
# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.1.0rc2 version (SVN $Revision: 3585 $)'


import gc3libs
from gc3libs.compat.functools import wraps
import gc3libs.exceptions
from gc3libs.quantity import Memory, kB, MB, GB
from gc3libs.quantity import Duration, hours, minutes, seconds
import gc3libs.utils


class LRMS(gc3libs.utils.Struct):
    """Base class for interfacing with a computing resource.

    The following construction parameters are also set as instance
    attributes.  All of them are mandatory, except `auth`.

    +---------------------+--------------+-------------------------------+
    |Attribute name       |Expected Type |Meaning                        |
    +=====================+==============+===============================+
    |`name`               |string        |A unique identifier for this   |
    |                     |              |resource, used for generating  |
    |                     |              |error message.                 |
    +---------------------+--------------+-------------------------------+
    |`architecture`       |set of        |Should contain one entry per   |
    |                     |`Run.Arch`    |each architecture              |
    |                     |values        |supported. Valid architecture  |
    |                     |              |values are constants in the    |
    |                     |              |`gc3libs.Run.Arch` class.      |
    +---------------------+--------------+-------------------------------+
    |`auth`               |string        |A `gc3libs.authentication.Auth`|
    |                     |              |instance that will be used to  |
    |                     |              |access the computational       |
    |                     |              |resource associated with this  |
    |                     |              |backend.  The default value    |
    |                     |              |`None` is used to mean that no |
    |                     |              |authentication credentials are |
    |                     |              |needed (e.g., access to the    |
    |                     |              |resource has been              |
    |                     |              |pre-authenticated) or is       |
    |                     |              |managed outside of GC3Pie).    |
    +---------------------+--------------+-------------------------------+
    |`max_cores`          |int           |Maximum number of CPU cores    |
    |                     |              |that GC3Pie can allocate on    |
    |                     |              |this resource.                 |
    +---------------------+--------------+-------------------------------+
    |`max_cores_per_job`  |int           |Maximum number of CPU cores    |
    |                     |              |that GC3Pie can allocate on    |
    |                     |              |this resource *for a single    |
    |                     |              |job*.                          |
    +---------------------+--------------+-------------------------------+
    |`max_memory_per_core`|Memory        |Maximum memory that            |
    |                     |              |GC3Pie can allocate to jobs on |
    |                     |              |this resource.  The value is   |
    |                     |              |*per core*, so the actual      |
    |                     |              |amount allocated to a single   |
    |                     |              |job is the value of this entry |
    |                     |              |multiplied by the number of    |
    |                     |              |cores requested by the job.    |
    +---------------------+--------------+-------------------------------+
    |`max_walltime`       |Duration      |Maximum wall-clock time        |
    |                     |              |that can be allotted           |
    |                     |              |to a single job running on this|
    |                     |              |resource.                      |
    +---------------------+--------------+-------------------------------+

    The above should be considered *immutable* attributes: they are
    specified at construction time and changed never after.

    The following attributes are instead dynamically provided (i.e.,
    defined by the `get_resource_status()` method or similar), thus
    can change over the lifetime of the object:

    ===================  =====
    Attribute name       Type
    ===================  =====
    free_slots           int
    user_run             int
    user_queued          int
    queued               int
    ===================  =====

    """
    def __init__(self, name,
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime, auth=None,
                 # additional arguments can set instance attributes
                 **extra_args):
        gc3libs.utils.Struct.__init__(self, **extra_args)

        self.name = str(name)
        self.updated = False

        if len(architecture) == 0:
            raise gc3libs.exceptions.InvalidType(
                "Empty value list for mandatory attribute 'architecture'")
        self.architecture = architecture

        self.max_cores = int(max_cores)
        self.max_cores_per_job = int(max_cores_per_job)
        assert isinstance(max_memory_per_core, Memory), \
            ("Expected `Memory` value for `max_memory_per_core`, "
             "got %s instead." % (type(max_memory_per_core),))
        self.max_memory_per_core = max_memory_per_core
        assert isinstance(max_walltime, Duration), \
            ("Expected `Duration` value for `max_walltime`, got %s instead."
             % (type(max_walltime),))
        self.max_walltime = max_walltime

        # see `authenticated` below
        self._auth_fn = auth

        gc3libs.log.info(
            "Computational resource '%s' initialized successfully.", self.name)

    @staticmethod
    def authenticated(fn):
        """
        Decorator: mark a function as requiring authentication.

        Each invocation of the decorated function causes a call to the
        `get` method of the authentication object (configured with the
        `auth` parameter to the class constructor).
        """
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            if self._auth_fn is not None:
                self._auth_fn()
            return fn(self, *args, **kwargs)
        return wrapper

    def cancel_job(self, app):
        """
        Cancel a running job.  If `app` is associated to a queued or
        running remote job, tell the execution middleware to cancel
        it.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.cancel_job()` called "
            "- this should have been defined in a derived class.")

    def free(self, app):
        """
        Free up any remote resources used for the execution of `app`.
        In particular, this should delete any remote directories and
        files.

        Call this method when `app.execution.state` is anything other
        than `TERMINATED` results in undefined behavior and will
        likely be the cause of errors later on.  Be cautious.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.free()` called "
            "- this should have been defined in a derived class.")

    def get_resource_status(self):
        """
        Update the status of the resource associated with this `LRMS`
        instance in-place.  Return updated `Resource` object.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.get_resource_status()` called "
            "- this should have been defined in a derived class.")

    def get_results(self, job, download_dir, overwrite=False):
        """
        Retrieve job output files into local directory `download_dir`
        (which must already exists).  Will not overwrite existing
        files, unless the optional argument `overwrite` is `True`.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.get_results()` called "
            "- this should have been defined in a derived class.")

    def update_job_state(self, app):
        """
        Query the state of the remote job associated with `app` and
        update `app.execution.state` accordingly.  Return the
        corresponding `Run.State`; see `Run.State` for more details.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.update_state()` called "
            "- this should have been defined in a derived class.")

    def submit_job(self, application, job):
        """
        Submit an `Application` instance to the configured
        computational resource; return a `gc3libs.Job` instance for
        controlling the submitted job.

        This method only returns if the job is successfully submitted;
        upon any failure, an exception is raised.

        *Note:*

          1. `job.state` is *not* altered; it is the caller's
             responsibility to update it.

          2. the `job` object may be updated with any information that
             is necessary for this LRMS to perform further operations on it.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.submit_job()` called "
            "- this should have been defined in a derived class.")

    def peek(self, app, remote_filename, local_file, offset=0, size=None):
        """
        Download `size` bytes (at offset `offset` from the start) from
        remote file `remote_filename` and write them into
        `local_file`.  If `size` is `None` (default), then snarf
        contents of remote file from `offset` unto the end.

        Argument `local_file` is either a local path name (string), or
        a file-like object supporting a `.write()` method.  If
        `local_file` is a path name, it is created if not existent,
        otherwise overwritten.

        Argument `remote_filename` is the name of a file in the remote job
        "sandbox".

        Any exception raised by operations will be passed through.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.peek()` called "
            "- this should have been defined in a derived class.")

    def validate_data(self, data_file_list=None):
        """
        Return True if the list of files is expressed in one of the
        file transfer protocols the LRMS supports.

        Return False otherwise.
        """
        raise NotImplementedError(
            "Abstract method 'LRMS.validate_data()' called "
            "- this should have been defined in a derived class.")

    def close(self):
        """
        Implement gracefully close on LRMS dependent resources
        e.g. transport
        """
        raise NotImplementedError(
            "Abstract method 'LRMS.close()' called "
            "- this should have been defined in a derived class.")


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="__init__",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   grosetta.py -- Front-end script for submitting ROSETTA jobs to SMSCG.
#
#   Copyright (C) 2010, 2011, 2012 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting ROSETTA jobs to SMSCG.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``grosetta --help`` for program usage instructions.
"""
__version__ = '2.1.0rc2 version (SVN $Revision$)'
# summary of user-visible changes
__changelog__ = """
  2012-02-28:
    * Use option ``-R`` to request a specific release of Rosetta.
  2011-03-28:
    * No longer require any input file on the command line: e.g., ``grosetta -l``
      will just list the status of jobs in the current session.  On the other hand,
      if the FLAGS file is given, then given must be also at least one input file.
  2010-12-20:
    * Initial release, forking off the old `grosetta`/`gdocking` sources.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'



# workaround Issue 95, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == '__main__':
    import grosetta
    grosetta.GRosettaScript().run()


# stdlib imports
import os
import os.path
import sys

# interface to Gc3libs
import gc3libs
from gc3libs.application.rosetta import RosettaApplication
from gc3libs.cmdline import SessionBasedScript, positive_int



## the script class

class GRosettaScript(SessionBasedScript):
    """
Run MiniRosetta on the specified INPUT files and fetch OUTPUT files
back from the execution machine; if OUTPUT is omitted, all '*.pdb',
'*.sc' and '*.fasc' files are retrieved.  Several instances can be run in
parallel, depending on the '-P' and '-p' options.

The `grosetta` command keeps a record of jobs (submitted, executed and
pending) in a session file (set name with the '-s' option); at each
invocation of the command, the status of all recorded jobs is updated,
output from finished jobs is collected, and a summary table of all
known jobs is printed.  New jobs are added to the session if the number
of wanted decoys (option '-P') is raised.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; `grosetta` will delay submission
of newly-created jobs so that this limit is never exceeded.

Note: the list of INPUT and OUTPUT files must be separated by ':'
(on the shell command line, put a space before and after).
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            usage = "%(prog)s [options] FLAGSFILE INPUT ... [: OUTPUT ...]",
            version = __version__, # module version == script version
            application = RosettaApplication,
            )

    def setup_options(self):
        self.add_param("-P", "--total-decoys", dest="total_decoys",
                       type=positive_int, default=1,
                       metavar="NUM",
                       help="Compute NUM decoys per input file (default: %(default)s)."
                       )
        self.add_param("-p", "--decoys-per-job", dest="decoys_per_job",
                       type=positive_int, default=1,
                       metavar="NUM",
                       help="Compute NUM decoys in a single job (default: %(default)s)."
                       " This parameter should be tuned so that the running time"
                       " of a single job does not exceed the maximum wall-clock time."
                       )
        self.add_param("-x", "--protocol", dest="protocol",
                       type=str, default="minirosetta.static",
                       metavar="PROTOCOL",
                       help="Run the specified Rosetta protocol/application; default: %(default)s")
        self.add_param("-R", "--release",
                       type=str, dest="rosetta_release", default="3.1",
                       metavar="NAME",
                       help="Numerical suffix to identify which version of Rosetta should be requested."
                       " For example: '-e 20110622' will run rosetta-svn20110622."
                       " (default: %(default)s)"
                       )

    def parse_args(self):
        self.instances_per_file = self.params.total_decoys
        self.instances_per_job = self.params.decoys_per_job
        self.extra['number_of_decoys_to_create'] = self.params.decoys_per_job

        # parse positional arguments
        args = self.params.args
        if len(args) == 0:
            # no args given, so no new jobs added to the session
            self.flags_file = 'PLEASE GIVE FLAGS FILE ON THE COMMAND LINE'
            inputs = [ ]
            self.outputs = [ ]
        else:
            # if FLAGS is given, then we want at least one input file
            if len(args) < 2:
                raise RuntimeError("Incorrect usage; please run '%s --help' to read instructions." % self.name)
            try:
                self.flags_file = args[0]
                del args[0]
                if ':' in args:
                    separator = args.index(':')
                    inputs = args[:separator]
                    self.outputs = args[(separator+1):]
                else:
                    inputs = args
                    self.outputs = [ ]
            except:
                raise RuntimeError("Incorrect usage; please run '%s --help' to read instructions." % self.name)

            # make flags file path absolute
            if not os.path.isabs(self.flags_file):
                self.flags_file = os.path.abspath(self.flags_file)
            if not os.path.exists(self.flags_file):
                raise RuntimeError("Flags file '%s' does not exist." % self.flags_file)
            self.log.info("Using flags file '%s'", self.flags_file)

        # massage input file list to have only absolute paths
        inputs_ = [ ]
        for path in inputs:
            if not os.path.exists(path):
                self.log.error("Cannot access input path '%s' - aborting.", path)
                raise RuntimeError("Cannot access input path '%s' - aborting.", path)
            else:
                # make paths absolute
                if not os.path.isabs(path):
                    path = os.path.abspath(path)
                inputs_.append(path)
        self.inputs = inputs_
        #self.log.debug("Gathered input files: '%s'" % str.join("', '", inputs))

    def new_tasks(self, extra):
        ## compute number of decoys already being computed in this session
        decoys = 0
        for task in self.session:
            start, end = task.jobname.split('--')
            decoys += int(end) - int(start)
        self.log.debug("Total no. of decoys already scheduled for computation: %d", decoys)

        # add jobs to the session, until we are computing the specified number of decoys
        # XXX: if the number of requested decoys is lowered, we should cancel jobs!
        if decoys < self.params.total_decoys:
            if decoys > 0:
                self.log.info("Already computing %d decoys, requested %d more.",
                              decoys, self.params.total_decoys - decoys)
            # create new jobs and add them to session
            for nr in range(decoys, self.params.total_decoys, self.params.decoys_per_job):
                jobname = ("%d--%d"
                           % (nr, min(self.params.total_decoys,
                                      nr + self.params.decoys_per_job - 1)))
                # yield new job to construct to `self._main()`
                yield (
                    jobname, RosettaApplication,
                    # args
                    (self.params.protocol, self.params.rosetta_release, self.inputs, self.outputs),
                    # kwargs
                    {'arguments':[ '-out:nstruct', str(self.params.decoys_per_job) ],
                     'flags_file':self.flags_file,
                     },
                    )



#! /usr/bin/env python
#
"""
"""
#
__version__ = '$Revision$'
__author__ = 'Sergio Maffioletti <sergio.maffioletti@gc3.uzh.ch>'
# summary of user-visible changes
__changelog__ = """
  2011-06-27:
    * Defined ABCApplication and basic SessionBasedScript
"""
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == '__main__':
    import cchem_gfit_abc_workflow
    cchem_gfit_abc_workflow.ABC_Workflow().run()


# stdlib imports
import os
import os.path
import sys

## interface to Gc3libs
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, existing_file, executable_file
from gc3libs.workflow import SequentialTaskCollection, ParallelTaskCollection


#######################################################################################

# path UML and APPPOT
ABC_UML_IMAGE="gsiftp://idgc3grid01.uzh.ch/local_repo/ABC/abc.img"
ABC_APPOTRUN="gsiftp://idgc3grid01.uzh.ch/local_repo/ABC/abc_appotrun.sh"
ABC_RUN="/home/acostantini/workflow/script/abc.sh"



#######################################################################################


#
# class MainSequentialABC --> 2 steps
#
class MainSequentialABC(SequentialTaskCollection):
        def __init__(self, name, ABC_UML_IMAGE, output, g3cfile, dimensions, fortran_pes, inputfilelist_abc, **extra_args):

                self.inputfilelist_abc = inputfilelist_abc
                self.output_folder = output
                self.extra_args = extra_args
                self.name = name
# first step --> __init__ source compiling

                first_task = Gfit3C_ABC_uml_Application(abc_uml_image_file=ABC_UML_IMAGE, output_folder=output, g3c_input_file=g3cfile, dimension_file=dimensions, surface_file=fortran_pes, **extra_args)
                SequentialTaskCollection.__init__(self, name, [first_task])

# second step --> __init__ ABC execution (parallel)

        def next(self, done):

                if done == 0:

                        first_task = self.tasks[done]

                        abc_executable = os.path.join(first_task.output_dir, "abc.x") #first_task.outputs["abc.x"].path)

                        second_task = ParallelABC(ABC_RUN, abc_executable, self.inputfilelist_abc, first_task.output_dir, **self.extra_args)
                        self.add(second_task)
                        return Run.State.RUNNING

                else:
                        return Run.State.TERMINATED

#######################################################################################


#
# class ParallelABC --> ABC execution (parallel)
#
class ParallelABC(ParallelTaskCollection):
        def __init__(self, executable, abc_executable, inputfilelist_abc, output_folder, **extra_args):

                parallel_task = []
                for input_file in inputfilelist_abc:
                        name = "ABC_execution_" + os.path.basename(input_file)

                        parallel_task.append(ABC_Application(executable, abc_executable, input_file, output_folder, **extra_args))

                ParallelTaskCollection.__init__(self, name, parallel_task)

        def terminated(self):
                self.execution.returncode = (0, 0)




#######################################################################################




#
# class Gfit3C_ABC_uml_Application --> compile on uml -->  ABC binary
#
class Gfit3C_ABC_uml_Application(Application):
    application_name = 'gfit3c_abc_uml'
    def __init__(self, abc_uml_image_file, output_folder, g3c_input_file=None, dimension_file=None, surface_file=None, **extra_args):
        inputs = [(abc_uml_image_file, "abc.img")]

        if surface_file:
                inputs.append((surface_file,os.path.basename(surface_file)))
                abc_prefix = os.path.basename(surface_file)
        elif g3c_input_file and dimension_file:
                inputs.append((g3c_input_file, os.path.basename(g3c_input_file)))
                inputs.append((dimension_file,"dimensions"))
                abc_prefix = os.path.basename(g3c_input_file)
        else:
                raise gc3libs.exceptions.InvalidArgument("Missing critical argument surface file [%s], g3c_file [%s], dimension_file [%s]" % (surface_file, g3c_input_file, dimension_file))

        extra_args['tags'] = ["TEST/APPPOT-0"]
        extra_args['output_dir'] = os.path.join(output_folder,"abc."+abc_prefix)

        gc3libs.Application.__init__(self,
                                        executable = "$APPPOT_STARTUP",
                                        arguments = ["--apppot", "abc.cow,abc.img"],
                                        inputs = inputs,
                                        outputs = [ ("abc.x", "abc."+abc_prefix), (os.path.basename(abc_prefix).split(".g3c")[0]+"_log.tgz",
                                        os.path.basename(abc_prefix).split(".g3c")[0]+"_log.tgz")],
                                        join = True,
                                        stdout = os.path.basename(abc_prefix).split(".g3c")[0]+".log",
                                        **extra_args
                                        )




#######################################################################################




# class ABC_Application --> single ABC run
#
class ABC_Application(Application):
    application_name = 'abc'
    def __init__(self, executable, abc_executable, input_file, output_folder, **extra_args):


        extra_args['output_dir'] = os.path.join(output_folder, os.path.basename(input_file))

        gc3libs.Application.__init__(self,
                                executable = os.path.basename(executable),
                                arguments = [os.path.basename(input_file)],
                                executables = [abc_executable],
                                inputs = [(abc_executable, os.path.basename(abc_executable)), (input_file, os.path.basename(input_file)), (executable, os.path.basename(executable))],
                                outputs = [],
                                # output_dir = os.path.join(output_folder,os.path.basename(input_file)),
                                join = True,
                                stdout = os.path.basename(input_file)+".out",
                                **extra_args
                                )




#######################################################################################



#
# SessionBasedScript__ ABC_Workflow
#
class ABC_Workflow(SessionBasedScript):
    """
    Main control of the ABC execution workflow.
    Depending on the input parameters, ABC executin is preceeded by ABC binary compilation
    Scan the specified INPUTDIR directories recursively for '.d' files,
    and submit a ABC job for each input file found; job progress is
    monitored and, when a job is done, its output file is retrieved back
    """
    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            #application = Gfit3C_ABC_uml_Application,
            input_filename_pattern = '*.d'
            )


    def setup_options(self):

# inputs file definitions
        self.add_param("--g3c",
                       action="store", dest="g3cfile",
                       type=existing_file, default=None,
                       help="G3C input file.")

        self.add_param("--dim",
                       action="store", dest="dimensions",
                       type=existing_file, default=None,
                       help="Surface file."
                       )
        self.add_param("--exec",
                       action="store", dest="abc_exec",
                       type=executable_file, default=None,
                       help="ABC binary file."
                       )
        self.add_param("--pes", action="store", dest="fortran_pes",
                       type=existing_file, default=None,
                       help="FORTRAN file for the potential energy surface."
                       )
        return


    def parse_args(self):
        self.inputfolder_abc = self.params.args[0]
        self.output_folder = self.params.output

# possible actions depending from input files

    def new_tasks(self, extra):

         extra_args = extra.copy()

# inizializzo la lista di files

         self.inputfilelist_abc = self._search_for_input_files([self.inputfolder_abc])


# dimensions+gc3file --> compiling gfit + abc
         if self.params.dimensions and self.params.g3cfile:
             name = "Gfit3C_"+str(os.path.basename(self.params.g3cfile))
# yield for MainSequential_ABC
             yield (name, MainSequentialABC, [
                     name,
                     ABC_UML_IMAGE,
                     #self.apppot_run,
                     #ABC_APPOTRUN,
                     self.params.output,
                     self.params.g3cfile,
                     self.params.dimensions,
                     self.params.fortran_pes,
                     self.inputfilelist_abc,
                     ], extra_args) # MainSequentialABC(ABC_URL_IMAGE, self.apppot_run, self.params.out, ..., **extra_args)

# pes file --> compiling abc
#         elif self.params.fortran_pes:
#             name = "ABC_"+str(os.path.basename(self.params.fortran_pes))
# yield for MainSequential_ABC
#             yield (name, MainSequentialABC, [
#                     name,
#                    ABC_UML_IMAGE,
#                    #self.apppot_run,
#                     #ABC_APPOTRUN,
#                     self.params.output,
#                    self.params.g3cfile,
#                    self.params.dimensions,
#                     self.params.fortran_pes,
#                     ], extra_args)

# abc executable --> abc execution in parallel
         elif self.params.abc_exec:
             name = "ABC_execution_"+str(os.path.basename(self.params.abc_exec))
# yield for ParallelABC
             yield (name, ParallelABC, [
                     ABC_RUN,
                     self.params.abc_exec,
                     self.inputfilelist_abc,
                     os.path.join(self.params.output, os.path.basename(self.params.abc_exec)),
                     ], extra_args)

         else :
                raise gc3libs.exceptions.InvalidUsage("Invalid use of the command line")
#######################################################################################



#! /usr/bin/env python
#
"""
ForwardPremium-specific methods and overloads.
"""
# Copyright (C) 2011, 2012 University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

__version__ = '$Revision$'
__author__ = 'Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
# summary of user-visible changes
__changelog__ = """

"""
__docformat__ = 'reStructuredText'

import gc3libs.debug
import gc3libs.application.apppot
import re, os
import numpy as np
from supportGc3 import lower, flatten, str2tuple, getIndex, extractVal, str2vals
from supportGc3 import format_newVal, update_parameter_in_file, safe_eval, str2mat, mat2str, getParameter
from paraLoop import paraLoop

from gc3libs import Application, Run
import shutil

import logbook, sys
from supportGc3 import wrapLogger

logger = wrapLogger(loggerName = __name__ + 'logger', streamVerb = 'DEBUG', logFile = __name__ + '.log')

class idRiskApplication(Application):

    application_name = 'idrisk'

    _invalid_chars = re.compile(r'[^_a-zA-Z0-9]+', re.X)

    def __init__(self, executable, arguments, inputs, outputs, output_dir, **extra_args):
        Application.__init__(self, executable, arguments, inputs, outputs, output_dir, **extra_args)

    def fetch_output_error(self, ex):

        if self.execution.state == Run.State.TERMINATING:
        # do notify task/main application that we're done
        # ignore error, let's continue
            self.execution.state = Run.State.TERMINATED
            logger.debug('fetch_output_error occured... continuing')
            if self.persistent_id:
                logger.debug('jobid: %s exception: %s' % (self.persistent_id, str(ex)))
            else:
                logger.debug('info: %s exception: %s' % (self.info, str(ex)))
            return None
        else:
        # non-terminal state, pass on error
            return ex

    # def submit_error(self, ex):
    #     logger.debug('submit_error occured... continuing')

    #     try:
    #         if self.lrms_jobid:
    #             logger.debug('jobid: %s info: %s exception: %s' % (self.lrms_jobid, self.info, str(ex)))
    #         else:
    #             logger.debug('info: %s exception: %s' % (self.info, str(ex)))
    #     except AttributeError:
    #         logger.debug('no `lrms_jobid` hence submission didnt happen')
    #     return None


    def terminated(self):
        """
        Analyze the retrieved output, with a threefold purpose:

        - set the exit code based on whether there is a
          `simulation.out` file containing a value for the
          `FamaFrenchBeta` parameter;

        - parse the output file `simulation.out` (if any) and set
          attributes on this object based on the values stored there:
          e.g., if the `simulation.out` file contains the line
          ``FamaFrenchBeta: 1.234567``, then set `self.FamaFrenchBeta
          = 1.234567`.  Attribute names are gotten from the labels in
          the output file by translating any invalid character
          sequence to a single `_`; e.g. ``Avg. hB`` becomes `Avg_hB`.

        - work around a bug in ARC where the output is stored in a
          subdirectory of the output directory.
        """
        output_dir = self.output_dir
        # if files are stored in `output/output/`, move them one level up
        if os.path.isdir(os.path.join(output_dir, 'output')):
            wrong_output_dir = os.path.join(output_dir, 'output')
            for entry in os.listdir(wrong_output_dir):
                dest_entry = os.path.join(output_dir, entry)
                if os.path.isdir(dest_entry):
                    # backup with numerical suffix
                    gc3libs.utils.backup(dest_entry)
                os.rename(os.path.join(wrong_output_dir, entry), dest_entry)
        # set the exitcode based on postprocessing the main output file
        simulation_out = os.path.join(output_dir, 'simulation.out')
        if os.path.exists(simulation_out):
            self.execution.exitcode = 0
        else:
            # no `simulation.out` found, signal error
            self.execution.exitcode = 2



class idRiskApppotApplication(idRiskApplication, gc3libs.application.apppot.AppPotApplication):
    _invalid_chars = re.compile(r'[^_a-zA-Z0-9]+', re.X)

    def __init__(self, executable, arguments, inputs, outputs, output_dir, **extra_args):
        print 'extra_args = %s' % extra_args
        gc3libs.application.apppot.AppPotApplication.__init__(self, executable, arguments, inputs, outputs, output_dir, **extra_args)



#! /usr/bin/env python
#
"""
Tests for the cmdline module
"""
# Copyright (C) 2012, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

import os
import shutil
import tempfile
import re

import cli.test
from nose.tools import assert_true

import gc3libs.cmdline
import gc3libs.session

class TestScript(cli.test.FunctionalTest):
    def __init__(self, *args, **extra_args):
        cli.test.FunctionalTest.__init__(self, *args, **extra_args)
        self.scriptdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'scripts')


    def setUp(self):
        cli.test.FunctionalTest.setUp(self)
        CONF_FILE="""
[auth/dummy]
type = ssh
username = dummy

[resource/localhost]
enabled = true
auth = dummy
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 8
max_walltime = 8
max_cores = 2
architecture = x86_64
override = False
resourcedir = %s
"""
        (fd, self.cfgfile) = tempfile.mkstemp()
        self.resourcedir = self.cfgfile + '.d'
        f = os.fdopen(fd, 'w')
        f.write(CONF_FILE % self.resourcedir)
        f.close()

    def tearDown(self):
        os.remove(self.cfgfile)
        shutil.rmtree(self.resourcedir)
        cli.test.FunctionalTest.tearDown(self)

    def test_simplescript(self):
        """Test a very simple script based on `SessionBasedScript`:class:

        The script is found in ``gc3libs/tests/scripts/simplescript.py``
        """
        # XXX: WARNING: This will only work if you have a "localhost"
        # section in gc3pie.conf!!!
        result = self.run_script('python',
                                 os.path.join(self.scriptdir, 'simplescript.py'),
                                 '-C', '1',
                                 '-s', 'TestOne',
                                 '--config-files', self.cfgfile,
                                 '-r', 'localhost')

        assert_true(re.match('.*TERMINATED\s+3/3\s+\(100.0+%\).*', result.stdout, re.S))

        # FIXME: output dir should be inside session dir
        session_dir = os.path.join(self.env.base_path, 'TestOne')
        assert_true(
            os.path.isdir(
                os.path.join(self.env.base_path, 'SimpleScript.out.d')
                )
            )
        assert_true(
            os.path.isfile(
                os.path.join(self.env.base_path, 'SimpleScript.out.d', 'SimpleScript.stdout')
                )
            )

        assert_true(
            os.path.isdir(
                os.path.join(self.env.base_path, 'SimpleScript.out2.d')
                )
            )
        assert_true(
            os.path.isfile(
                os.path.join(self.env.base_path, 'SimpleScript.out2.d', 'SimpleScript.stdout')
                )
            )

        assert_true(
            os.path.isdir(session_dir)
            )

        assert_true(
            os.path.isfile(
                os.path.join(session_dir, gc3libs.session.Session.INDEX_FILENAME, )
                )
            )

        assert_true(
            os.path.isfile(
                os.path.join(session_dir, gc3libs.session.Session.STORE_URL_FILENAME, )
                )
            )


## main: run tests

if "__main__" == __name__:
    import nose
    nose.runmodule()



#! /usr/bin/env python
#
"""
Test interaction with the SLURM batch-queueing system.
"""
# Copyright (C) 2011, 2012, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

import datetime
import os
import tempfile


from nose.plugins.skip import SkipTest
from nose.tools import assert_equal

import gc3libs
from gc3libs.backends.slurm import count_jobs
import gc3libs.core
import gc3libs.config
from gc3libs.quantity import Memory, kB, MB, GB, Duration, seconds, minutes, hours

from faketransport import FakeTransport


def correct_submit(jobid=123):
    return (
        # command exit code
        0,
        # stdout
        ("Submitted batch job %d\n" % jobid),
        # stderr
        '')

def sacct_no_accounting(jobid=123):
    return (
        # command exitcode
        1,
        # stdout
        '',
        # stderr
        'SLURM accounting storage is disabled\n')

def squeue_pending(jobid=123):
    # squeue --noheader --format='%i|%T|%r' -j $jobid
    return (
        # command exitcode
        0,
        # stdout
        ('%d:PENDING:Resources' % jobid),
        # stderr
        '')

def squeue_running(jobid=123):
    # squeue --noheader --format='%i|%T|%u|%U|%r|%R' -j $jobid
    return (
        # command exitcode
        0,
        # stdout
        ('%d:RUNNING:None' % jobid),
        # stderr
        '')

def squeue_recently_completed(jobid=123):
    # squeue --noheader --format='%i|%T|%u|%U|%r|%R' -j $jobid
    return (
        # command exitcode
        0,
        # stdout
        '',
        # stderr
        '')

def squeue_notfound(jobid=123):
    # squeue --noheader --format='%i|%T|%u|%U|%r|%R' -j $jobid
    return (
        # command exitcode
        1,
        # stdout
        '',
        # stderr
        'slurm_load_jobs error: Invalid job id specified')


def scancel_success(jobid=123):
    return (0, "", "")

def scancel_notfound(jobid=123):
    return (
        # command exitcode
        1,
        # stdout
        '',
        # stderr
        ('scancel: error: Kill job error on job id %d: Invalid job id specified' % jobid))

def scancel_permission_denied(jobid=123):
    return (
        # command exitcode (yes, it's really 0!)
        0,
        # stdout
        "",
        # stderr
        "scancel: error: Kill job error on job id %d: Access/permission denied\n" % jobid)


def sacct_done_bad_timestamps(jobid=123):
    #    $ sudo sacct --noheader --parsable --format jobid,ncpus,cputimeraw,elapsed,submit,start,end,exitcode,maxrss,maxvmsize,totalcpu -j 19
    return (
        # command exitcode (yes, it's really 0!)
        0,
        # stdout
        """
123|0:0|1|00:01:06|00:00:00|2012-09-24T10:48:34|2012-09-24T10:47:28|2012-09-24T10:48:34|||
123.0|0:0|1|00:01:05|00:00:00|2012-09-24T10:47:29|2012-09-24T10:47:29|2012-09-24T10:48:34|0|0|
        """.strip(),
        # stderr
        "")


def sacct_done_parallel(jobid=123):
    # sample gotten from J.-G. Piccinali, CSCS
    return (
        # command exitcode
        0,
        # stdout
        """
123|0:0|64|00:00:23|00:01.452|2012-09-04T11:18:06|2012-09-04T11:18:24|2012-09-04T11:18:47|||
123.batch|0:0|1|00:00:23|00:01.452|2012-09-04T11:18:24|2012-09-04T11:18:24|2012-09-04T11:18:47|7884K|49184K|
        """.strip(),
        # stderr
        "")


def sacct_done_relative_timestamps(jobid=123):
    # sample gotten from J.-G. Piccinali, CSCS using SLURM_TIME_FORMAT=relative
    return (
        # command exitcode
        0,
        # stdout
        """
123|0:0|64|00:00:23|00:01.452|4 Sep 11:18|4 Sep 11:18|4 Sep 11:18|||
123.batch|0:0|1|00:00:23|00:01.452|4 Sep 11:18|4 Sep 11:18|4 Sep 11:18|7884K|49184K|
        """.strip(),
        # stderr
        "")


def sacct_no_accounting(jobid=123):
    return (
        # command exitcode
        1,
        # stdout
        "",
        # stderr
        "SLURM accounting storage is disabled\n")


#import gc3libs.Run.State as State
State = gc3libs.Run.State


class FakeApp(gc3libs.Application):
    def __init__(self):
        gc3libs.Application.__init__(
            self,
            arguments = ['/bin/hostname'],               # mandatory
            inputs = [],                  # mandatory
            outputs = [],                 # mandatory
            output_dir = "./fakedir",    # mandatory
            stdout = "stdout.txt",
            stderr = "stderr.txt",
            requested_cores = 1,)


class TestBackendSlurm(object):

    CONF="""
[resource/example]
type=slurm
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64
enabled=True

[auth/ssh]
type=ssh
username=NONEXISTENT
"""
    def setUp(self):
        (fd, self.tmpfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        f.write(TestBackendSlurm.CONF)
        f.close()

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(self.tmpfile)

        self.core = gc3libs.core.Core(self.cfg)

        self.backend = self.core.get_backend('example')
        self.backend.transport = FakeTransport()
        self.transport = self.backend.transport
        # basic responses
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['scancel'] = scancel_notfound()

    def tearDown(self):
        os.remove(self.tmpfile)

    def test_slurm_basic_workflow(self):
        app = FakeApp()

        # Succesful submission:
        self.transport.expected_answer['sbatch'] = correct_submit()
        self.core.submit(app)
        assert_equal(app.execution.state, State.SUBMITTED)

        # Update state. We would expect the job to be SUBMITTED
        self.transport.expected_answer['squeue'] = squeue_pending()
        self.core.update_job_state(app)
        assert_equal(app.execution.state, State.SUBMITTED)

        # Update state. We would expect the job to be RUNNING
        self.transport.expected_answer['squeue'] = squeue_running()
        self.core.update_job_state(app)
        assert_equal(app.execution.state, State.RUNNING)

        # Job done. qstat doesn't find it, tracejob should.
        self.transport.expected_answer['squeue'] = squeue_recently_completed()
        # XXX: alternatively:
        #self.transport.expected_answer['squeue'] = squeue_notfound()
        #self.transport.expected_answer['sacct'] = sacct_done()
        self.core.update_job_state(app)
        assert_equal(app.execution.state, State.TERMINATING)


    def test_parse_sacct_output_parallel(self):
        """Test `sacct` output with a successful parallel job."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = correct_submit()
        self.core.submit(app)
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['sacct'] = sacct_done_parallel()
        self.core.update_job_state(app)
        assert_equal(app.execution.state, State.TERMINATING)

        job = app.execution
        # common job reporting values (see Issue 78)
        assert_equal(job.cores,           64)
        assert_equal(job.exitcode,        0)
        assert_equal(job.signal,          0)
        assert_equal(job.duration,        23*seconds)
        assert_equal(job.max_used_memory, 49184*kB)
        assert_equal(job.used_cpu_time,   1.452*seconds)
        # SLURM-specific values
        assert_equal(job.slurm_max_used_ram, 7884*kB)
        assert_equal(job.slurm_submission_time,
                     datetime.datetime(year=2012, month=9, day=4, hour=11, minute=18, second=6))
        assert_equal(job.slurm_start_time,
                     datetime.datetime(year=2012, month=9, day=4, hour=11, minute=18, second=24))
        assert_equal(job.slurm_completion_time,
                     datetime.datetime(year=2012, month=9, day=4, hour=11, minute=18, second=47))


    def test_parse_sacct_output_bad_timestamps(self):
        """Test `sacct` output with out-of-order timestamps."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = correct_submit()
        self.core.submit(app)
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['sacct'] = sacct_done_bad_timestamps()
        self.core.update_job_state(app)
        assert_equal(app.execution.state, State.TERMINATING)

        job = app.execution
        # common job reporting values (see Issue 78)
        assert_equal(job.exitcode,        0)
        assert_equal(job.signal,          0)
        assert_equal(job.duration,        66*seconds)
        assert_equal(job.max_used_memory, 0*kB)
        assert_equal(job.used_cpu_time,   0*seconds)
        # SLURM-specific values
        assert_equal(job.slurm_max_used_ram, 0*kB)
        assert_equal(job.slurm_submission_time,
                     datetime.datetime(year=2012, month=9, day=24, hour=10, minute=47, second=28))
        assert_equal(job.slurm_start_time,
                     datetime.datetime(year=2012, month=9, day=24, hour=10, minute=47, second=29))
        assert_equal(job.slurm_completion_time,
                     datetime.datetime(year=2012, month=9, day=24, hour=10, minute=48, second=34))


    def test_cancel_job1(self):
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = correct_submit()
        self.core.submit(app)

        self.transport.expected_answer['scancel'] = scancel_success()
        self.core.kill(app)
        assert_equal(app.execution.state, State.TERMINATED)

    def test_cancel_job2(self):
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = correct_submit()
        self.core.submit(app)
        assert_equal(app.execution.state, State.SUBMITTED)

        self.transport.expected_answer['scancel'] = scancel_permission_denied()
        self.core.kill(app)
        assert_equal(app.execution.state, State.TERMINATED)

    def test_get_command(self):
        assert_equal(self.backend.sbatch,   ['sbatch'])
        assert_equal(self.backend._sacct,   'sacct')
        assert_equal(self.backend._scancel, 'scancel')
        assert_equal(self.backend._squeue,  'squeue')


def test_get_command_alternate():
    (fd, tmpfile) = tempfile.mkstemp()
    f = os.fdopen(fd, 'w+')
    f.write("""
[auth/ssh]
type=ssh
username=NONEXISTENT

[resource/example]
# mandatory stuff
type=slurm
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64

# alternate command paths
sbatch  = /usr/local/bin/sbatch --constraint=gpu
sacct   = /usr/local/sbin/sacct
scancel = /usr/local/bin/scancel # comments are ignored!
squeue  = /usr/local/bin/squeue
""")
    f.close()

    cfg = gc3libs.config.Configuration()
    cfg.merge_file(tmpfile)
    os.remove(tmpfile)
    b = cfg.make_resources()['example']

    assert_equal(b.sbatch,   ['/usr/local/bin/sbatch', '--constraint=gpu'])
    assert_equal(b._sacct,    '/usr/local/sbin/sacct')
    assert_equal(b._scancel,  '/usr/local/bin/scancel')
    assert_equal(b._squeue,   '/usr/local/bin/squeue')

def test_count_jobs():
    squeue_stdout = '''474867:PENDING:first_user:21913:Priority:(Priority)
474870:PENDING:first_user:21913:Priority:(Priority)
475753:PENDING:first_user:21913:Resources:(Resources)
475747:PENDING:first_user:21913:Resources:(Resources)
475751:PENDING:first_user:21913:Priority:(Priority)
474723:RUNNING:third_user:20345:None:nid0[0002,0005,0013,0017,0022-0023,0028-0029,0040,0046,0050,0058-0059,0061,0066,0068-0069,0081,0091,0100,0110,0122-0123,0132-0133,0186-0187,0193,0212-0213,0230-0231,0234-0235,0240,0276-0277,0280-0281,0294-0295,0298-0299,0302,0314,0325,0332-0333,0335,0337,0341,0344-0345,0352-0353,0358-0359,0370-0371,0386-0387,0429,0444-0445,0450-0451,0466-0467,0492,0508,0528-0529,0544-0545,0551,0562-0563,0578-0579,0608,0628,0636-0637,0642-0644,0651,0654-0655,0658,0670-0673,0684-0685,0698-0699,0709,0735,0744-0745,0771,0776-0777,0780-0781,0802-0803,0818,0822-0823,0828,0832-0833,0840,0885,0894,0906-0907,0909,0946,0948,0972-0973,1011,1022,1024-1025,1052,1059,1074,1098,1100-1103,1106,1122,1132-1133,1137,1139,1141-1143,1164-1165,1169,1198-1199,1202,1232-1233,1254-1255,1262,1274-1275,1304,1315,1324-1325,1334-1335,1356-1359,1394-1395,1420-1421,1424,1455,1459,1488,1492-1493,1519]
475738:RUNNING:first_user:21913:None:nid00[136-137]
475744:RUNNING:first_user:21913:None:nid00[182-183]
475438:RUNNING:second_user:21239:None:nid00[686,720-721,751]
475440:RUNNING:second_user:21239:None:nid00[363,875,916-917]
475448:RUNNING:second_user:21239:None:nid0[0306-0307,1206-1207]
475450:RUNNING:second_user:21239:None:nid0[1263-1265,1296]
475452:RUNNING:second_user:21239:None:nid0[0656,0687,1070,1105]
475736:RUNNING:first_user:21913:None:nid0[0041,1512]
475651:RUNNING:second_user:21239:None:nid00[026,217,867,924]
475726:RUNNING:first_user:21913:None:nid00[742-743]
'''
    # first user has 4 running jobs and 5 queued ones
    R, Q, r, q = gc3libs.backends.slurm.count_jobs(squeue_stdout, 'first_user')
    assert_equal(R, 11)
    assert_equal(Q, 5)
    assert_equal(r, 4)
    assert_equal(q, 5)
    # second user has 6 running jobs and no queued ones
    R, Q, r, q = gc3libs.backends.slurm.count_jobs(squeue_stdout, 'second_user')
    assert_equal(R, 11)
    assert_equal(Q, 5)
    assert_equal(r, 6)
    assert_equal(q, 0)
    # third user has only 1 running job
    R, Q, r, q = gc3libs.backends.slurm.count_jobs(squeue_stdout, 'third_user')
    assert_equal(R, 11)
    assert_equal(Q, 5)
    assert_equal(r, 1)
    assert_equal(q, 0)


if __name__ == "__main__":
    import nose
    nose.runmodule()



#! /usr/bin/env python
#
"""
Job control on PBS/Torque clusters (possibly connecting to the
front-end via SSH).
"""
# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.1.0rc2 version (SVN $Revision$)'


import datetime
import os
import posixpath
import random
import re
import sys
import tempfile
import time

from gc3libs.compat._collections import defaultdict

from gc3libs import log, Run
from gc3libs.backends import LRMS
import gc3libs.exceptions
from gc3libs.quantity import Memory, kB, MB, GB
from gc3libs.quantity import Duration, seconds, minutes, hours
import gc3libs.utils as utils  # first, to_bytes
from gc3libs.utils import same_docstring_as

import transport

import batch


## data for parsing PBS commands output

# regexps for extracting relevant strings

_qsub_jobid_re = re.compile(r'(?P<jobid>\d+.*)', re.I)

_qstat_line_re = re.compile(
    r'^(?P<jobid>\d+)[^\d]+\s+'
    '(?P<jobname>[^\s]+)\s+'
    '(?P<username>[^\s]+)\s+'
    '(?P<time_used>[^\s]+)\s+'
    '(?P<state>[^\s]+)\s+'
    '(?P<queue>[^\s]+)')

_tracejob_queued_re = re.compile(
    '(?P<submission_time>\d+/\d+/\d+\s+\d+:\d+:\d+)\s+.\s+'
    'Job Queued at request of .*job name =\s*(?P<job_name>[^,]+),'
    '\s+queue =\s*(?P<queue>[^,]+)')

_tracejob_run_re = re.compile(
    '(?P<running_time>\d+/\d+/\d+\s+\d+:\d+:\d+)\s+.\s+'
    'Job Run at request of .*')

_tracejob_last_re = re.compile(
    '(?P<end_time>\d+/\d+/\d+\s+\d+:\d+:\d+)\s+.'
    '\s+Exit_status=(?P<exit_status>\d+)\s+'
    'resources_used.cput=(?P<used_cpu_time>[^ ]+)\s+'
    'resources_used.mem=(?P<mem>[^ ]+)\s+'
    'resources_used.vmem=(?P<used_memory>[^ ]+)\s+'
    'resources_used.walltime=(?P<used_walltime>[^ ]+)')

# convert data to GC3Pie internal format


def _to_memory(val):
    """
    Convert a memory quantity as it appears in the PBS logs to GC3Pie
    `Memory` value.

    Examples::

      >>> _to_memory('44kb') == 44*kB
      True
      >>> _to_memory('12mb') == 12*MB
      True
      >>> _to_memory('2gb') == 2*GB
      True
      >>> _to_memory('1024') == Memory(1024, Memory.B)
      True

    """
    # extract the `kb`, `mb`, etc. suffix, if there is one
    unit = val[-2:]
    # XXX: check that PBS uses base-2 units
    if unit == 'kb':
        return int(val[:-2])*Memory.kB
    elif unit == 'mb':
        return int(val[:-2])*Memory.MB
    elif unit == 'gb':
        return int(val[:-2])*Memory.GB
    elif unit == 'tb':
        return int(val[:-2])*Memory.TB
    else:
        if val[-1] == 'b':
            # XXX bytes
            val = int(val[:-1])
        else:
            # a pure number
            val = int(val)
        return int(val)*Memory.B


def _parse_asctime(val):
    try:
        # XXX: replace with datetime.strptime(...) in Python 2.5+
        return datetime.datetime(
            *(time.strptime(val, '%m/%d/%Y %H:%M:%S')[0:6]))
    except Exception, err:
        gc3libs.log.error(
            "Cannot parse '%s' as a PBS-format time stamp: %s: %s",
            val, err.__class__.__name__, str(err))
        return None


_tracejob_keyval_mapping = {
    # regexp group name
    # |               `Task.execution` attribute
    # |               |
    # |               |                       converter function
    # |               |                       |
    # |               |                       |
    #   ... common backend attrs (see Issue 78) ...
    'exit_status': ('exitcode', int),
    'used_cpu_time': ('used_cpu_time', Duration),
    'used_walltime': ('duration', Duration),
    'used_memory': ('max_used_memory', _to_memory),
    #   ... PBS-only attrs ...
    'mem': ('pbs_max_used_ram', _to_memory),
    'submission_time': ('pbs_submission_time', _parse_asctime),
    'running_time': ('pbs_running_time', _parse_asctime),
    'end_time': ('pbs_end_time', _parse_asctime),
    'queue': ('pbs_queue', str),
    'job_name': ('pbs_jobname', str), }


## code

def count_jobs(qstat_output, whoami):
    """
    Parse PBS/Torque's ``qstat`` output (as contained in string `qstat_output`)
    and return a quadruple `(R, Q, r, q)` where:

      * `R` is the total number of running jobs in the PBS/Torque cell
        (from any user);

      * `Q` is the total number of queued jobs in the PBS/Torque cell
        (from any user);

      * `r` is the number of running jobs submitted by user `whoami`;

      * `q` is the number of queued jobs submitted by user `whoami`
    """
    total_running = 0
    total_queued = 0
    own_running = 0
    own_queued = 0
    n = 0
    for line in qstat_output.split('\n'):
        # import pdb; pdb.set_trace()
        log.info("Output line: %s" % line)
        m = _qstat_line_re.match(line)
        if not m:
            continue
        if m.group('state') in ['R']:
            total_running += 1
            if m.group('username') == whoami:
                own_running += 1
        elif m.group('state') in ['Q']:
            total_queued += 1
            if m.group('username') == whoami:
                own_queued += 1
        log.info("running: %d, queued: %d" % (total_running, total_queued))

    return (total_running, total_queued, own_running, own_queued)


class PbsLrms(batch.BatchSystem):
    """
    Job control on PBS/Torque clusters (possibly by connecting via SSH
    to a submit node).
    """

    _batchsys_name = 'PBS/TORQUE'

    def __init__(self, name,
                 # this are inherited from the base LRMS class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 auth,  # ignored if `transport` is 'local'
                 # these are inherited from `BatchSystem`
                 frontend, transport,
                 accounting_delay=15,
                 # these are specific to this backend
                 queue=None,
                 **extra_args):

        # init base class
        batch.BatchSystem.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth,
            frontend, transport, accounting_delay=accounting_delay,
            **extra_args)

        # backend-specific setup
        self.queue = queue
        self.qsub = self._get_command_argv('qsub')

        # PBS/TORQUE commands
        self._qdel = self._get_command('qdel')
        self._qstat = self._get_command('qstat')
        self._tracejob = self._get_command('tracejob')

    def _parse_submit_output(self, output):
        return self.get_jobid_from_submit_output(output, _qsub_jobid_re)

    def _submit_command(self, app):
        qsub_argv, app_argv = app.qsub_pbs(self)
        if self.queue is not None:
            qsub_argv += ['-d', '.', '-q', ('%s' % self.queue)]
        return (str.join(' ', qsub_argv), str.join(' ', app_argv))

    def _stat_command(self, job):
        return "%s %s | grep ^%s" % (
            self._qstat, job.lrms_jobid, job.lrms_jobid)

    def _acct_command(self, job):
        return "%s %s" % (self._tracejob, job.lrms_jobid)

    def _parse_stat_output(self, stdout):
        # check that passed object obeys contract

        # parse `qstat` output
        job_status = stdout.split()[4]
        jobstatus = dict()
        log.debug("translating PBS/Torque's `qstat` code "
                  "'%s' to gc3libs.Run.State", job_status)
        if job_status in ['Q', 'W']:
            jobstatus['state'] = Run.State.SUBMITTED
        elif job_status in ['R']:
            jobstatus['state'] = Run.State.RUNNING
        elif job_status in ['S', 'H', 'T'] or 'qh' in job_status:
            jobstatus['state'] = Run.State.STOPPED
        elif job_status in ['C', 'E']:
            jobstatus['state'] = Run.State.TERMINATING
        else:
            jobstatus['state'] = Run.State.UNKNOWN

        return jobstatus

    def _parse_acct_output(self, stdout):
        jobstatus = {}
        for line in stdout.split('\n'):
            # XXX: make a list of the three regexps and loop over it
            match = _tracejob_queued_re.match(line)
            if match:
                for key, value in match.groupdict().iteritems():
                    attr, conv = _tracejob_keyval_mapping[key]
                    jobstatus[attr] = conv(value)
                continue
            match = _tracejob_run_re.match(line)
            if match:
                for key, value in match.groupdict().iteritems():
                    attr, conv = _tracejob_keyval_mapping[key]
                    jobstatus[attr] = conv(value)
                continue
            match = _tracejob_last_re.match(line)
            if match:
                for key, value in match.groupdict().iteritems():
                    attr, conv = _tracejob_keyval_mapping[key]
                    jobstatus[attr] = conv(value)
                break
        return jobstatus

    def _cancel_command(self, jobid):
        return ("%s %s" % (self._qdel, jobid))

    @same_docstring_as(LRMS.get_resource_status)
    @LRMS.authenticated
    def get_resource_status(self):
        self.updated = False
        try:
            self.transport.connect()

            _command = ('%s -a' % self._qstat)
            log.debug("Running `%s`...", _command)
            exit_code, qstat_stdout, stderr \
                = self.transport.execute_command(_command)
            if exit_code != 0:
                # cannot continue
                raise gc3libs.exceptions.LRMSError(
                    "PBS backend failed executing '%s':"
                    " exit code: %d; stdout: '%s', stderr: '%s'"
                    % (_command, exit_code, stdout, stderr))

            log.debug("Computing updated values for total/available slots ...")
            (total_running, self.queued, self.user_run, self.user_queued) \
                = count_jobs(qstat_stdout, self._username)
            self.total_run = total_running
            self.free_slots = -1
            self.used_quota = -1

            log.info("Updated resource '%s' status:"
                     " free slots: %d,"
                     " total running: %d,"
                     " own running jobs: %d,"
                     " own queued jobs: %d,"
                     " total queued jobs: %d",
                     self.name,
                     self.free_slots,
                     self.total_run,
                     self.user_run,
                     self.user_queued,
                     self.queued,
                     )
            return self

        except Exception, ex:
            # self.transport.close()
            log.error("Error querying remote LRMS, see debug log for details.")
            log.debug("Error querying LRMS: %s: %s",
                      ex.__class__.__name__, str(ex), exc_info=True)
            raise


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="sge",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
This module provides a generic BatchSystem class from which all
batch-like backends should inherit.
"""
# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.1.0rc2 version (SVN $Revision$)'


from getpass import getuser
import os
import posixpath
import random
import shlex
import stat
import sys
import tempfile
import time

import gc3libs
from gc3libs import log, Run
from gc3libs.backends import LRMS
from gc3libs.utils import ifelse, same_docstring_as
import gc3libs.backends.transport

# Define some commonly used functions

# FIXME: (Riccardo?) thinks this function is completely wrong and only
# exists to support GAMESS' ``qgms``, which does not allow users to
# specify the name of STDOUT/STDERR files.  When we have a standard
# flexible submission mechanism for all applications, we should remove
# it!


def generic_filename_mapping(jobname, jobid, file_name):
    """
    Map STDOUT/STDERR filenames (as recorded in `Application.outputs`)
    to commonly used default STDOUT/STDERR file names (e.g.,
    ``<jobname>.o<jobid>``).
    """
    try:
        return {('%s.out' % jobname): ('%s.o%s' % (jobname, jobid)),
                ('%s.err' % jobname): ('%s.e%s' % (jobname, jobid)),
                # FIXME: the following is definitely GAMESS-specific
                ('%s.cosmo' % jobname): ('%s.o%s.cosmo' % (jobname, jobid)),
                ('%s.dat' % jobname): ('%s.o%s.dat' % (jobname, jobid)),
                ('%s.inp' % jobname): ('%s.o%s.inp' % (jobname, jobid)),
                ('%s.irc' % jobname): ('%s.o%s.irc' % (jobname, jobid))
                }[file_name]
    except KeyError:
        return file_name


def _make_remote_and_local_path_pair(transport, job, remote_relpath,
                                     local_root_dir, local_relpath):
    """
    Return list of (remote_path, local_path) pairs corresponding to
    """
    # see https://github.com/fabric/fabric/issues/306 about why it is
    # correct to use `posixpath.join` for remote paths (instead of
    # `os.path.join`)
    remote_path = posixpath.join(job.ssh_remote_folder,
                                 generic_filename_mapping(job.lrms_jobname,
                                                          job.lrms_jobid,
                                                          remote_relpath))
    local_path = os.path.join(local_root_dir, local_relpath)
    if transport.isdir(remote_path):
        # recurse, accumulating results
        result = []
        for entry in transport.listdir(remote_path):
            result += _make_remote_and_local_path_pair(
                transport, job,
                posixpath.join(remote_relpath, entry),
                local_path, entry)
        return result
    else:
        return [(remote_path, local_path)]


class BatchSystem(LRMS):
    """
    Base class for backends dealing with a batch-queue system (e.g.,
    PBS/TORQUE, Grid Engine, etc.)

    This is an abstract class, that you should subclass in order to
    interface with a given batch queuing system. (Remember to call
    this class' constructor in the derived class ``__init__`` method.)

    """

    _batchsys_name = 'batch queueing system'
    """
    A human-readable identifier for the batch queueing system
    class/type (e.g., PBS, LSF, etc.).
    """

    def __init__(self, name,
                 # this are inherited from the base LRMS class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 auth,  # ignored if `transport` is 'local'
                 # these are specific to the this backend
                 frontend, transport, keyfile=None,
                 accounting_delay=15,
                 **extra_args):

        # init base class
        LRMS.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth, **extra_args)

        # backend-specific setup
        self.frontend = frontend
        if transport == 'local':
            self.transport = gc3libs.backends.transport.LocalTransport()
            self._username = getuser()
        elif transport == 'ssh':
            auth = self._auth_fn()
            self._username = auth.username
            self.transport = gc3libs.backends.transport.SshTransport(
                frontend, username=self._username, keyfile=keyfile)
        else:
            raise gc3libs.exceptions.TransportError(
                "Unknown transport '%s'" % transport)

    def get_jobid_from_submit_output(self, output, regexp):
        """Parse the output of the submission command. Regexp is
        provided by the caller. """
        for line in output.split('\n'):
            match = regexp.match(line)
            if match:
                return match.group('jobid')
        raise gc3libs.exceptions.InternalError(
            "Could not extract jobid from qsub output '%s'"
            % qsub_output.rstrip())

    def _get_command_argv(self, name, default=None):
        """
        Return an *argv*-style array for invoking command `name`.

        The command name is looked up in this resource's configuration
        parameters, and, if found, the associated string is split into
        words (according to the normal POSIX shell rules) and this
        list of words is returned::

          | >>> b = BatchSystem(..., bsub='/usr/local/bin/bsub -R lustre')
          | >>> b._get_command('bsub')
          | ['/usr/local/bin/bsub', '-R', 'lustre']

        Otherwise, if no configuration parameter by name `name` is
        found, then second argument `default` is returned as the sole
        member of the *argv*-array::

          | >>> b = BatchSystem(...)
          | >>> b._get_command('foo', 'bar')
          | ['bar']

        If `default` is `None`, then the value of `name` is used
        instead::

          | >>> b._get_command('baz')
          | ['baz']

        """
        if default is None:
            default = name
        # lookup the command name in the resource config parameters;
        # return it unchanged as a default
        cmd = self.get(name, default)
        # return argv-style array
        return shlex.split(cmd, comments=True)

    def _get_command(self, name, default=None):
        """
        Return an command-line (string) for invoking command `name`.

        The command name is looked up in this resource's configuration
        parameters, and, if found, the associated string is returned::

          | >>> b = BatchSystem(..., bsub='/usr/local/bin/bsub -R lustre')
          | >>> b._get_command('bsub')
          | '/usr/local/bin/bsub -R lustre'

        Otherwise, if no configuration parameter by name `name` is
        found, then second argument `default` is returned::
        member of the *argv*-array::

          | >>> b = BatchSystem(...)
          | >>> b._get_command('foo', 'bar')
          | 'bar'

        If `default` is `None`, then the value of `name` is used
        instead::

          | >>> b._get_command('baz')
          | 'baz'

        """
        if default is None:
            default = name
        # lookup the command name in the resource config parameters;
        # return it unchanged as a default
        argv = self._get_command_argv(name, default)
        return str.join(' ', argv)

    def _submit_command(self, app):
        """This method returns a string containing the command to
        issue to submit the job."""
        raise NotImplementedError(
            "Abstract method `_submit_command()` called - "
            "this should have been defined in a derived class.")

    def _parse_submit_output(self, stdout):
        """This method will parse the output of the submit command and
        return the jobid of the submitted job."""
        raise NotImplementedError(
            "Abstract method `parse_submit_output()` called - "
            "this should have been defined in a derived class.")

    def _stat_command(self, job):
        """This method returns a string containing the command to
        issue to get status information about a job."""
        raise NotImplementedError(
            "Abstract method `_stat_command()` called - "
            "this should have been defined in a derived class.")

    def _parse_stat_output(self, stdout):
        """This method will parse the output of the stat command and
        return the current status of the job. The return value will be
        a dictionary which will be used to update job's information.

        The only expected key is `state`, which must be a valid
        `Run.State` state.
        """
        raise NotImplementedError(
            "Abstract method `_parse_stat_output()` called - "
            "this should have been defined in a derived class.")

    def _acct_command(self, job):
        """This method returns a string containing the command to
        issue to get accounting information about the `job`.

        It is usually called only if the _stat_command() fails.
        """
        raise NotImplementedError(
            "Abstract method `_acct_command()` called - "
            "this should have been defined in a derived class.")

    def _parse_acct_output(self, stdout):
        """This method will parse the output of the acct command and
        return a dictionary containing infos about the
        job. `BatchSystem` class does not make any assumption about
        the keys contained in the dictionary.
        """
        raise NotImplementedError(
            "Abstract method `_parse_acct_output()` called - "
            "this should have been defined in a derived class.")

    def _cancel_command(self, jobid):
        """This method returns a string containing the command to
        issue to delete the job identified by `jobid`
        """
        raise NotImplementedError(
            "Abstract method `_cancel_command()` called -"
            " this should have been defined in a derived class.")

    def _get_prepost_scripts(self, app, scriptnames):
        script_txt = []
        for script in scriptnames:
            if script not in self:
                gc3libs.log.debug(
                    "%s script not defined for resource %s", script, self.name)
                continue
            if not os.path.isfile(self[script]):
                gc3libs.log.debug(
                    "%s script points to file '%s', which does"
                    " not exist - ignoring.", script, self[script])
                continue
            gc3libs.log.debug("Adding %s file `%s` to the submission script"
                              % (script, self[script]))
            script_file = open(self[script])
            script_txt.append("\n# %s file `%s` BEGIN\n"
                              % (script, self[script]))
            script_txt.append(script_file.read())
            script_txt.append("\n# %s file END\n" % script)
            script_file.close()
        return str.join("", script_txt)

    def get_prologue_script(self, app):
        """
        This method will get the prologue script(s) for the `app`
        application and will return a string which contains the
        contents of the script(s) merged together.
        """
        prologues = ['prologue', app.application_name+'_prologue']
        return self._get_prepost_scripts(app, prologues)

    def get_epilogue_script(self, app):
        """
        This method will get the epilogue script(s) for the `app`
        application and will return a string which contains the
        contents of the script(s) merged together.
        """
        epilogues = ['epilogue', app.application_name+'epilogue']
        return self._get_prepost_scripts(app, epilogues)

    @LRMS.authenticated
    def submit_job(self, app):
        """This method will create a remote directory to store job's
        sandbox, and will copy the sandbox in there.
        """
        job = app.execution
        # Create the remote directory.
        try:
            self.transport.connect()

            cmd = "mkdir -p $HOME/.gc3pie_jobs;" \
                " mktemp -d $HOME/.gc3pie_jobs/lrms_job.XXXXXXXXXX"
            log.info("Creating remote temporary folder: command '%s' " % cmd)
            exit_code, stdout, stderr = self.transport.execute_command(cmd)
            if exit_code == 0:
                ssh_remote_folder = stdout.split('\n')[0]
            else:
                raise gc3libs.exceptions.LRMSError(
                    "Failed executing command '%s' on resource '%s';"
                    " exit code: %d, stderr: '%s'."
                    % (cmd, self.name, exit_code, stderr))
        except gc3libs.exceptions.TransportError, x:
            raise
        except:
            raise
                # Copy the input file to remote directory.
        for local_path, remote_path in app.inputs.items():
            remote_path = os.path.join(ssh_remote_folder, remote_path)
            remote_parent = os.path.dirname(remote_path)
            try:
                if remote_parent not in ['', '.']:
                    log.debug("Making remote directory '%s'",
                              remote_parent)
                    self.transport.makedirs(remote_parent)
                log.debug("Transferring file '%s' to '%s'",
                          local_path.path, remote_path)
                self.transport.put(local_path.path, remote_path)
                # preserve execute permission on input files
                if os.access(local_path.path, os.X_OK):
                    self.transport.chmod(remote_path, 0755)
            except:
                log.critical(
                    "Copying input file '%s' to remote cluster '%s' failed",
                    local_path.path, self.frontend)
                raise

        if app.arguments[0].startswith('./'):
            gc3libs.log.debug("Making remote path '%s' executable.",
                              app.arguments[0])
            self.transport.chmod(os.path.join(ssh_remote_folder,
                                              app.arguments[0]), 0755)

        try:
            sub_cmd, aux_script = self._submit_command(app)
            if aux_script != '':
                # create temporary script name
                # XXX: The `uuid` module is available from Py 2.5 onwards
                script_filename = ('./script.%x.sh'
                                   % random.randint(0, sys.maxint))
                # save script to a temporary file and submit that one instead
                local_script_file = tempfile.NamedTemporaryFile()
                local_script_file.write('#!/bin/sh\n')
                # Add preamble file
                prologue = self.get_prologue_script(app)
                if prologue:
                    local_script_file.write(prologue)

                local_script_file.write(aux_script)

                # Add epilogue files
                epilogue = self.get_epilogue_script(app)
                if epilogue:
                    local_script_file.write(epilogue)

                local_script_file.flush()
                # upload script to remote location
                self.transport.put(
                    local_script_file.name,
                    os.path.join(ssh_remote_folder, script_filename))
                # set execution mode on remote script
                self.transport.chmod(
                    os.path.join(ssh_remote_folder, script_filename), 0755)
                # cleanup
                local_script_file.close()
                if os.path.exists(local_script_file.name):
                    os.unlink(local_script_file.name)
            else:
                # we still need a script name even if there is no
                # script to submit
                script_filename = ''

            # Submit it
            exit_code, stdout, stderr = self.transport.execute_command(
                "/bin/sh -c 'cd %s && %s %s'"
                % (ssh_remote_folder, sub_cmd, script_filename))

            if exit_code != 0:
                raise gc3libs.exceptions.LRMSError(
                    "Failed executing command '%s' on resource '%s';"
                    " exit code: %d, stderr: '%s'."
                    % (cmd, self.name, exit_code, stderr))

            jobid = self._parse_submit_output(stdout)
            log.debug('Job submitted with jobid: %s', jobid)

            job.execution_target = self.frontend

            job.lrms_jobid = jobid
            job.lrms_jobname = jobid
            try:
                if app.jobname:
                    job.lrms_jobname = app.jobname
            except:
                pass

            if 'stdout' in app:
                job.stdout_filename = app.stdout
            else:
                job.stdout_filename = '%s.o%s' % (job.lrms_jobname, jobid)
            if app.join:
                job.stderr_filename = job.stdout_filename
            else:
                if 'stderr' in app:
                    job.stderr_filename = app.stderr
                else:
                    job.stderr_filename = '%s.e%s' % (job.lrms_jobname, jobid)
            job.history.append('Submitted to %s @ %s, got jobid %s'
                               % (self._batchsys_name, self.name, jobid))
            job.history.append("Submission command output:\n"
                               "  === stdout ===\n%s"
                               "  === stderr ===\n%s"
                               "  === end ===\n"
                               % (stdout, stderr), 'pbs', 'qsub')
            job.ssh_remote_folder = ssh_remote_folder

            return job

        except:
            log.critical(
                "Failure submitting job to resource '%s' - "
                "see log file for errors", self.name)
            raise

    @same_docstring_as(LRMS.update_job_state)
    @LRMS.authenticated
    def update_job_state(self, app):
        try:
            job = app.execution
            job.lrms_jobid
        except AttributeError, ex:
            # `job` has no `lrms_jobid`: object is invalid
            raise gc3libs.exceptions.InvalidArgument(
                "Job object is invalid: %s" % str(ex))

        try:
            self.transport.connect()
            cmd = self._stat_command(job)
            log.debug("Checking remote job status with '%s' ..." % cmd)
            exit_code, stdout, stderr = self.transport.execute_command(cmd)
            if exit_code == 0:
                jobstatus = self._parse_stat_output(stdout)
                job.update(jobstatus)

                job.state = jobstatus.get('state', Run.State.UNKNOWN)
                if job.state == Run.State.UNKNOWN:
                    log.warning(
                        "Unknown batch job status,"
                        " setting GC3Pie job state to `UNKNOWN`")

                if 'exit_status' in jobstatus:
                    job.exitcode = int(jobstatus['exit_status'])
                    # XXX: we should set the `signal` part accordingly
                    job.signal = 0

                # SLURM's `squeue` command exits with code 0 if the
                # job ID exists in the database (i.e., a job with that
                # ID has been run) but prints no output.  In this
                # case, we need to continue and examine the accounting
                # command output to get the termination status etc.
                if job.state != Run.State.TERMINATING:
                    return job.state

            # In some batch systems, jobs disappear from qstat
            # output as soon as they are finished. In these cases,
            # we have to check some *accounting* command to check
            # the exit status.
            cmd = self._acct_command(job)
            if cmd:
                log.debug(
                    "The `qstat`/`bjobs` command returned no job information;"
                    " trying with '%s' instead ..." % cmd)
                exit_code, stdout, stderr = self.transport.execute_command(cmd)
                if exit_code == 0:
                    jobstatus = self._parse_acct_output(stdout)
                    job.update(jobstatus)
                    if 'exitcode' in jobstatus:
                        job.returncode = int(jobstatus['exitcode'])
                        job.state = Run.State.TERMINATING
                    return job.state

            # No *stat command and no *acct command returned
            # correctly.
            try:
                if (time.time() - job.stat_failed_at) > self.accounting_delay:
                    # accounting info should be there, if it's not
                    # then job is definitely lost
                    log.critical(
                        "Failed executing remote command: '%s';"
                        "exit status %d", cmd, exit_code)
                    log.debug(
                        "  remote command returned stdout: '%s'" % stdout)
                    log.debug(
                        "  remote command returned stderr: '%s'" % stderr)
                    raise gc3libs.exceptions.LRMSError(
                        "Failed executing remote command: '%s'; exit status %d"
                        % (cmd, exit_code))
                else:
                    # do nothing, let's try later...
                    return job.state
            except AttributeError:
                # this is the first time `qstat` fails, record a
                # timestamp and retry later
                job.stat_failed_at = time.time()

        except Exception, ex:
            log.error("Error in querying Batch resource '%s': %s: %s",
                      self.name, ex.__class__.__name__, str(ex))
            raise
        # If we reach this point it means that we don't actually know
        # the current state of the job.
        job.state = Run.State.UNKNOWN
        return job.state

    @same_docstring_as(LRMS.peek)
    @LRMS.authenticated
    def peek(self, app, remote_filename, local_file, offset=0, size=None):
        job = app.execution
        assert 'ssh_remote_folder' in job, \
            "Missing attribute `ssh_remote_folder` on `Job` instance" \
            " passed to `PbsLrms.peek`."

        if size is None:
            size = sys.maxint

        _filename_mapping = generic_filename_mapping(
            job.lrms_jobname, job.lrms_jobid, remote_filename)
        _remote_filename = os.path.join(
            job.ssh_remote_folder, _filename_mapping)

        try:
            self.transport.connect()
            remote_handler = self.transport.open(
                _remote_filename, mode='r', bufsize=-1)
            remote_handler.seek(offset)
            data = remote_handler.read(size)
        except Exception, ex:
            log.error("Could not read remote file '%s': %s: %s",
                      _remote_filename, ex.__class__.__name__, str(ex))

        try:
            local_file.write(data)
        except (TypeError, AttributeError):
            output_file = open(local_file, 'w+b')
            output_file.write(data)
            output_file.close()
        log.debug('... Done.')

    @same_docstring_as(LRMS.validate_data)
    def validate_data(self, data_file_list):
        """
        Supported protocols: file
        """
        for url in data_file_list:
            if not url.scheme in ['file']:
                return False
        return True

    @same_docstring_as(LRMS.cancel_job)
    @LRMS.authenticated
    def cancel_job(self, app):
        job = app.execution
        try:
            self.transport.connect()
            cmd = self._cancel_command(job.lrms_jobid)
            exit_code, stdout, stderr = self.transport.execute_command(cmd)
            if exit_code != 0:
                # It is possible that 'qdel' fails because job has
                # been already completed thus the cancel_job behaviour
                # should be to
                log.error(
                    "Failed executing remote command '%s'; exit status %d",
                    cmd, exit_code)
                log.debug("  remote command returned stdout '%s'", stdout)
                log.debug("  remote command returned stderr '%s'", stderr)
                if exit_code == 127:
                    # no such command
                    raise gc3libs.exceptions.LRMSError(
                        "Cannot execute remote command '%s'" % cmd)

            return job

        except:
            log.critical('Failure in checking status')
            raise

    @same_docstring_as(LRMS.free)
    @LRMS.authenticated
    def free(self, app):

        job = app.execution
        try:
            self.transport.connect()
            self.transport.remove_tree(job.ssh_remote_folder)
        except:
            log.warning("Failed removing remote folder '%s': %s: %s",
                        job.ssh_remote_folder, sys.exc_info()[0],
                        sys.exc_info()[1])
        return

    @same_docstring_as(LRMS.get_results)
    @LRMS.authenticated
    def get_results(self, app, download_dir, overwrite=False):
        if app.output_base_url is not None:
            raise gc3libs.exceptions.UnrecoverableDataStagingError(
                "Retrieval of output files to non-local destinations"
                " is not supported (yet).")

        job = app.execution
        try:
            self.transport.connect()
            # Make list of files to copy, in the form of (remote_path,
            # local_path) pairs.  This entails walking the
            # `Application.outputs` list to expand wildcards and
            # directory references.
            stageout = list()
            for remote_relpath, local_url in app.outputs.iteritems():
                local_relpath = local_url.path
                if remote_relpath == gc3libs.ANY_OUTPUT:
                    remote_relpath = ''
                    local_relpath = ''
                stageout += _make_remote_and_local_path_pair(
                    self.transport, job, remote_relpath, download_dir,
                    local_relpath)

            # copy back all files, renaming them to adhere to the
            # ArcLRMS convention
            log.debug("Downloading job output into '%s' ...", download_dir)
            for remote_path, local_path in stageout:
                log.debug("Downloading remote file '%s' to local file '%s'",
                          remote_path, local_path)
                if (overwrite
                        or not os.path.exists(local_path)
                        or os.path.isdir(local_path)):
                    log.debug("Copying remote '%s' to local '%s'"
                              % (remote_path, local_path))
                    # ignore missing files (this is what ARC does too)
                    self.transport.get(remote_path, local_path,
                                       ignore_nonexisting=True)
                else:
                    log.info("Local file '%s' already exists;"
                             " will not be overwritten!",
                             local_path)

            return  # XXX: should we return list of downloaded files?

        except:
            raise

    @same_docstring_as(LRMS.validate_data)
    @LRMS.authenticated
    def close(self):
        self.transport.close()


if "__main__" == __name__:
    import doctest
    doctest.testmod(name="sge",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
'''
Created on Jun 5, 2012

@author: quandtan
'''

import os
import sys
from ruffus import *
from cStringIO import StringIO
from subprocess import Popen
from subprocess import PIPE
from applicake.framework.runner import IniFileRunner, ApplicationRunner
from applicake.framework.runner import CollectorRunner
from applicake.framework.runner import WrapperRunner
from applicake.applications.proteomics.openbis.generator import GuseGenerator
from applicake.applications.os.echo import Echo
from applicake.applications.commons.collector import GuseCollector
from applicake.applications.proteomics.searchengine.xtandem import Xtandem
from applicake.applications.proteomics.openbis.dss import Dss
from applicake.applications.proteomics.tpp.tandem2xml import Tandem2Xml
from applicake.applications.proteomics.tpp.xinteract import Xinteract
from applicake.applications.proteomics.tpp.interprophet import InterProphet
from applicake.applications.proteomics.openms.filehandling.idfileconverter import PepXml2IdXml
from applicake.applications.proteomics.openms.peptideproteinprocessing.falsediscoveryrate import FalseDiscoveryRate
from applicake.applications.proteomics.openms.peptideproteinprocessing.peptideindexer import PeptideIndexer
from applicake.applications.proteomics.openms.peptideproteinprocessing.idfilter import IdFilter
from applicake.applications.proteomics.openms.filehandling.fileconverter import Mzxml2Mzml
from applicake.applications.proteomics.openms.signalprocessing.peakpickerhighres import PeakPickerHighRes
from applicake.applications.proteomics.openms.quantification.featurefindercentroided import OrbiLessStrict
from applicake.applications.proteomics.sybit.pepxml2csv import Pepxml2Csv
from applicake.applications.proteomics.sybit.fdr2probability import Fdr2Probability
from applicake.applications.proteomics.tpp.proteinprophet import ProteinProphet
from applicake.applications.proteomics.sybit.protxml2spectralcount import ProtXml2SpectralCount
from applicake.applications.proteomics.sybit.protxml2modifications import ProtXml2Modifications
from applicake.applications.proteomics.sybit.protxml2openbis import ProtXml2Openbis
from applicake.applications.proteomics.openbis.dropbox import Copy2Dropbox,\
    Copy2IdentDropbox
from applicake.applications.commons.inifile import Unifier
from applicake.framework.interfaces import IApplication, IWrapper

cwd = None

#helper function
def wrap(applic,  input_file_name, output_file_name,opts=None):
    argv = ['', '-i', input_file_name, '-o', output_file_name]
    if opts is not None:
        argv.extend(opts)
    application = applic()
    if isinstance(application, IApplication):
        runner = ApplicationRunner()
        print 'use application runner'
    elif isinstance(application, IWrapper):
        runner = WrapperRunner()
    else:
        raise Exception('could not identfy [%s]' % applic.__name__)   
    application = applic()
    exit_code = runner(argv, application)
    if exit_code != 0:
        raise Exception("[%s] failed [%s]" % (applic.__name__, exit_code))


def execute(command):
    p = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)           
    output, error = p.communicate()                                                                                                                                                                           
    out_stream = StringIO(output)
    err_stream = StringIO(error)


def setup():
    cwd = '.'
    os.chdir(cwd)
    execute("find . -type d -iname '[0-9]*' -exec rm -rf {} \;")
    execute('rm *.err')
    execute('rm *.out')
    execute('rm *.log')
    execute('rm *ini*')
#    execute('rm jobid.txt')
    execute('rm flowchart.*')   
    with open("input.ini", 'w+') as f:
        f.write("""BASEDIR = /cluster/scratch/malars/workflows
LOG_LEVEL = DEBUG
STORAGE = file
TEMPLATE = template.tpl
DATASET_DIR = /cluster/scratch/malars/datasets
DATASET_CODE = 20120124102254267-296925,20120124121656335-296961
DBASE = /cluster/scratch/malars/biodb/ex_sp/current/decoy/ex_sp_9606.fasta
DECOY_STRING = DECOY_
FRAGMASSERR = 0.4
FRAGMASSUNIT = Da
PRECMASSERR = 15
PRECMASSUNIT = ppm
MISSEDCLEAVAGE = 0
ENZYME = Trypsin
STATIC_MODS = Carbamidomethyl (C)
THREADS = 4
XTANDEM_SCORE = k-score
XINTERACT_ARGS = -dDECOY_ -OAPdlIw
IPROPHET_ARGS = MINPROB=0
FDR=0.01
SPACE = QUANDTAN
PROJECT = TEST
DROPBOX = /cluster/scratch/malars/drop-box_prot_ident
""" 
)      
       

@follows(setup)
@split("input.ini", "generate.ini_*")
def generator(input_file_name, notused_output_file_names):
    argv = ['', '-i', input_file_name, '--GENERATORS', 'generate.ini','-o','generator.ini','-l','DEBUG']
    runner = IniFileRunner()
    application = GuseGenerator()
    exit_code = runner(argv, application)
    if exit_code != 0:
        raise Exception("generator failed [%s]" % exit_code)
   
@transform(generator, regex("generate.ini_"), "dss.ini_")
def dss(input_file_name, output_file_name):  
    wrap(Dss,input_file_name, output_file_name,['--PREFIX', 'getmsdata'])


@transform(dss, regex("dss.ini_"), "xtandem.ini_")
def tandem(input_file_name, output_file_name):
    wrap(Xtandem,input_file_name, output_file_name,['--PREFIX', 'tandem.exe','-s','file','-l','DEBUG'])

@transform(tandem, regex("xtandem.ini_"), "xtandem2xml.ini_")
def tandem2xml(input_file_name, output_file_name):
    wrap(Tandem2Xml,input_file_name, output_file_name) 

@transform(tandem2xml, regex("xtandem2xml.ini_"), "xinteract.ini_")
def xinteract(input_file_name, output_file_name):
    wrap(Xinteract,input_file_name, output_file_name)  

   
@merge(xinteract, "collector.ini")
def collector(notused_input_file_names, output_file_name):
    argv = ['', '--COLLECTORS', 'xinteract.ini', '-o', output_file_name,'-s','file']
    runner = CollectorRunner()
    application = GuseCollector()
    exit_code = runner(argv, application)
    if exit_code != 0:
        raise Exception("[%s] failed [%s]" % ('collector',exit_code))    

@follows(collector)
def unifier():
    argv = ['', '-i', 'collector.ini', '-o','unifier.ini','-p','--UNIFIER_REDUCE']
    runner = IniFileRunner()
    application = Unifier()
    exit_code = runner(argv, application)
    if exit_code != 0:
        raise Exception("unifier [%s]" % exit_code) 

@follows(unifier)
def interprophet():
    wrap(InterProphet,'unifier.ini','interprophet.ini')   

@follows(interprophet)
def pepxml2csv():
    wrap(Pepxml2Csv,'interprophet.ini','pepxml2csv.ini')   

@follows(pepxml2csv)
def fdr2probability():
    wrap(Fdr2Probability,'pepxml2csv.ini','fdr2probability.ini')        

@follows(fdr2probability)
def proteinprophet():
    wrap(ProteinProphet,'fdr2probability.ini','proteinprophet.ini')

@follows(proteinprophet)
def protxml2spectralcount():
    wrap(ProtXml2SpectralCount,'proteinprophet.ini','protxml2spectralcount.ini')

@follows(protxml2spectralcount)
def protxml2modifications():
    wrap(ProtXml2Modifications,'protxml2spectralcount.ini','protxml2modifications.ini')

@follows(protxml2modifications)
def protxml2openbis():
    wrap(ProtXml2Openbis,'protxml2modifications.ini','protxml2openbis.ini') 

@follobws(protxml2openbis)
def copy2dropbox():
    argv = ['', '-i', 'protxml2openbis.ini', '-o','copy2dropbox.ini','-p']
    runner = IniFileRunner()
    application = Copy2IdentDropbox()
    exit_code = runner(argv, application)
    if exit_code != 0:
        raise Exception("unifier [%s]" % exit_code) 

pipeline_run([copy2dropbox])






#! /usr/bin/env python
"""Exceptions specific to the `gc3libs` package.

In addition to the exceptions listed here, `gc3libs`:mod: functions
try to use Python builtin exceptions with the same meaning they have
in core Python, namely:

* `TypeError` is raised when an argument to a function or method has an
  incompatible type or does not implement the required protocol (e.g.,
  a number is given where a sequence is expected).

* `ValueError`is  raised when an argument to a function or method has
  the correct type, but fails to satisfy other constraints in the
  function contract (e.g., a positive number is required, and `-1` is
  passed instead).

* `AssertionError` is raised when some internal assumption regarding
  state or function/method calling contract is violated.  Informally,
  this indicates a bug in the software.

"""
# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.1.0rc2 version (SVN $Revision: 3585 $)'


import gc3libs

## base error classes

class Error(Exception):
    """
    Base class for all error-level exceptions in GC3Pie.

    Generally, this indicates a non-fatal error: depending on the
    nature of the task, steps could be taken to continue, but users
    *must* be aware that an error condition occurred, so the message
    is sent to the logs at the ERROR level.

    Exceptions indicating an error condition after which the program
    cannot continue and should immediately stop, should use the
    `FatalError`:class: base class.
    """
    def __init__(self, msg, do_log=False):
        if do_log:
            gc3libs.log.error(msg)
        Exception.__init__(self, msg)


## mark errors as "Recoverable" (meaning that a retry a ta later time
## could succeed), or "Unrecoverable" (meaning there's no point in
## retrying).

class RecoverableError(Error):
    """
    Used to mark transient errors: retrying the same action at a later
    time could succeed.

    This exception should *never* be instanciated: it is only to be used
    in `except` clauses to catch "try again" situations.
    """
    pass


class UnrecoverableError(Error):
    """
    Used to mark permanent errors: there's no point in retrying the same
    action at a later time, because it will yield the same error again.

    This exception should *never* be instanciated: it is only to be used
    in `except` clauses to exclude "try again" situations.
    """
    pass


class FatalError(UnrecoverableError):
    """
    A fatal error: execution cannot continue and program should report
    to user and then stop.

    The message is sent to the logs at CRITICAL level
    when the exception is first constructed.

    This is the base class for all fatal exceptions.
    """
    def __init__(self, msg, do_log=True):
        if do_log:
            gc3libs.log.critical(msg)
        Exception.__init__(self, msg)


## derived exceptions

class AuthError(Error):
    """
    Base class for Auth-related errors.

    Should *never* be instanciated: create a specific error class
    describing the actual error condition.
    """
    pass

class RecoverableAuthError(AuthError, RecoverableError):
    pass

class UnrecoverableAuthError(AuthError, UnrecoverableError):
    pass


class ConfigurationError(FatalError):
    """
    Raised when the configuration file (or parts of it) could not be
    read/parsed.  Also used to signal that a required parameter is
    missing or has an unknown/invalid value.
    """
    pass


class DataStagingError(Error):
    """
    Base class for data staging and movement errors.

    Should *never* be instanciated: create a specific error class
    describing the actual error condition.
    """
    pass

class RecoverableDataStagingError(DataStagingError, RecoverableError):
    """
    Raised when transient problems with copying data to or from the
    remote execution site occurred.

    This error is considered to be transient (e.g., network
    connectivity interruption), so trying again at a later time could
    solve the problem.
    """
    pass

class UnrecoverableDataStagingError(DataStagingError, UnrecoverableError):
    """
    Raised when problems with copying data to or from the remote
    execution site occurred.
    """
    pass


class InputFileError(FatalError):
    """
    Raised when an input file is specified, which does not exist or
    cannot be read.
    """
    pass


class InternalError(Error):
    """
    Raised when some function cannot fulfill its duties, for reasons
    that do not depend on the library client code.  For instance, when
    a response string gotten from an external command cannot be parsed
    as expected.
    """
    pass


class InvalidArgument(Error, AssertionError): # XXX: should this be fatal?
    """
    Raised when the arguments passed to a function do not honor some
    required contract.  For instance, either one of two optional
    arguments must be provided, but none of them was.
    """
    pass

class InvalidType(InvalidArgument, TypeError):
    """
    A specialization of`InvalidArgument` for cases when the type of
    the passed argument does not match expectations.
    """
    pass

class InvalidValue(InvalidArgument, ValueError):
    """
    A specialization of`InvalidArgument` for cases when the type of
    the passed argument does not match expectations.
    """
    pass

class DuplicateEntryError(InvalidArgument):
    """
    Raised by `Application.__init__` if not all (local or remote)
    entries in the input or output files are distinct.
    """
    pass


class InvalidOperation(Error):
    """
    Raised when an operation is attempted, that is not considered
    valid according to the system state.  For instance, trying to
    retrieve the output of a job that has not yet been submitted.
    """
    pass


class InvalidResourceName(Error, ValueError):
    """
    Raised to signal that no computational resource with the given
    name is defined in the configuration file.
    """
    pass


class InvalidUsage(FatalError):
    """
    Raised when a command is not provided all required arguments on
    the command line, or the arguments do not match the expected
    syntax.

    Since the exception message is the last thing a user will see,
    try to be specific about what is wrong on the command line.
    """
    pass


class LoadError(Error):
    """
    Raised upon errors loading a job from the persistent storage.
    """
    pass


class LRMSError(Error):
    pass


class LRMSSubmitError(Error):
    pass

class LRMSSkipSubmissionToNextIteration(RecoverableError):
    """
    Tells the Engine to skip submission of a job until the next
    iteration.
    """
    pass

class NoConfigurationFile(FatalError):
    """
    Raised when the configuration file cannot be read (e.g., does not
    exist or has wrong permissions), or cannot be parsed (e.g., is
    malformed).
    """
    pass


class NoResources(Error):
    """
    Raised to signal that no resources are defined, or that none are
    compatible with the request.
    """
    # FIXME: should we have a separate `NoCompatibleResources` exception?
    pass


class OutputNotAvailableError(InvalidOperation):
    """
    Raised upon attempts to retrieve the output for jobs that are
    still in `NEW` or `SUBMITTED` state.
    """
    pass


class TaskError(Error):
    """
    Generic error condition in a `Task` object.
    """
    pass


class DetachedFromGridError(TaskError):
    """
    Raised when a method (other than :meth:`attach`) is called on
    a detached `Task` instance.
    """
    pass


class UnexpectedStateError(TaskError):
    """
    Raised by :meth:`Task.progress` when a job lands in `STOPPED`
    or `TERMINATED` state.
    """
    pass


class TransportError(Error):
    pass

class RecoverableTransportError(RecoverableError):
    pass

class UnrecoverableTransportError(UnrecoverableError):
    pass

class CopyError(TransportError):
    """
    Error copying a file from `source` to `destination.
    """
    def __init__(self, source, destination, ex):
        self.source = source
        self.destination = destination
        TransportError.__init__(
            self,
            "Could not copy '%s' to '%s': %s: %s"
            % (source, destination, ex.__class__.__name__, str(ex)))

class UnknownJob(Error, ValueError):
    """
    Raised when an operation is attempted on a task, which is
    unknown to the remote server or backend.
    """
    pass


class UnknownJobState(Error, AssertionError):
    """
    Raised when a job state is gotten from the Grid middleware, that
    is not handled by the GC3Libs code.  Might actually mean that
    there is a version mismatch between GC3Libs and the Grid
    middleware used.
    """
    pass

class ApplicationDescriptionError(FatalError):
    """
    Raised when the dumped description on a given Application produces something that the LRMS backend cannot process.
    As an example, for arc backends, this error is raised when the parsing of the Application's XRSL fails
    """
    pass


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="exceptions",
                    optionflags=doctest.NORMALIZE_WHITESPACE)
